/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Arith
@Title     = Arithmetic on two files
@Section   = Arithmetic
@Arguments = ifile1 ifile2 ofile
@Operators = add sub mul div min max

@EndModule


@BeginOperator_add

@Title     = Add two fields

@BeginDesciption
@BeginMath
o(t,x) = i_1(t,x) + i_2(t,x)
@EndMath
@EndDesciption

@EndOperator


@BeginOperator_sub

@Title     = Subtract two fields

@BeginDesciption
@BeginMath
o(t,x) = i_1(t,x) - i_2(t,x)
@EndMath
@EndDesciption

@EndOperator


@BeginOperator_mul

@Title     = Multiply two fields

@BeginDesciption
@BeginMath
o(t,x) = i_1(t,x) * i_2(t,x)
@EndMath
@EndDesciption

@EndOperator


@BeginOperator_div

@Title     = Divide two fields

@BeginDesciption
@BeginMath
o(t,x) = i_1(t,x) / i_2(t,x)
@EndMath
@EndDesciption

@EndOperator


@BeginOperator_min

@Title     = Minimum of two fields

@BeginDesciption
@BeginMath
o(t,x) = MIN(i_1(t,x), i_2(t,x))
@EndMath
@EndDesciption

@EndOperator


@BeginOperator_max

@Title     = Maximum of two fields

@BeginDesciption
@BeginMath
o(t,x) = MAX(i_1(t,x), i_2(t,x))
@EndMath
@EndDesciption

@EndOperator

@EndDoc
*/

#define  NOPERATORS  6

void *Arith(void *argument)
{
  static char func[] = "Arith";
  const int noperators = NOPERATORS;
  char *operators[NOPERATORS];
  int operatorID;
  int function[NOPERATORS];
  int streamID1, streamID2, streamID3;
  int gridsize;
  int nrecs, nrecs2, nvars = 0, nlev, recID;
  int tsID;
  int varID, levelID;
  int offset;
  int vlistID1, vlistID2, vlistID3;
  int lptype = 0;
  FIELD field1, field2;
  int **varnmiss2 = NULL;
  double **vardata2 = NULL;

  operators[0] = "add";  function[0] = func_add;
  operators[1] = "sub";  function[1] = func_sub;
  operators[2] = "mul";  function[2] = func_mul;
  operators[3] = "div";  function[3] = func_div;
  operators[4] = "min";  function[4] = func_min;
  operators[5] = "max";  function[5] = func_max;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));
  streamID2 = streamOpenRead(cdoStreamName(1));
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = streamInqVlist(streamID2);
  vlistID3 = vlistDuplicate(vlistID1);

  if ( vlistNrecs(vlistID2) == 1 && vlistNrecs(vlistID1) != 1 )
    lptype = 2;
  else
    vlistCompare(vlistID1, vlistID2, func_sft);
  
  if ( function[operatorID] == func_mul || function[operatorID] == func_div )
    nospec(vlistID1);

  streamID3 = streamOpenWrite(cdoStreamName(2), cdoFiletype());
  if ( streamID3 < 0 ) cdiError(streamID3, "Open failed on %s", cdoStreamName(2));

  streamDefVlist(streamID3, vlistID3);

  gridsize = vlistGridsizeMax(vlistID1);

  field1.ptr = (double *) malloc(gridsize*sizeof(double));
  field2.ptr = (double *) malloc(gridsize*sizeof(double));

  if ( cdoVerbose )
    cdoPrint("Number of timesteps: file1 %d, file2 %d", vlistNtsteps(vlistID1), vlistNtsteps(vlistID2));

  if ( vlistNtsteps(vlistID2) == 1 && vlistNtsteps(vlistID1) != 1 && lptype == 0 )
    {
      lptype = 1;
      nvars  = vlistNvars(vlistID2);
      vardata2  = (double **) malloc(nvars*sizeof(double *));
      varnmiss2 = (int **) malloc(nvars*sizeof(int *));
      for ( varID = 0; varID < nvars; varID++ )
	{
	  gridsize = gridInqSize(vlistInqVarGrid(vlistID2, varID));
	  nlev     = zaxisInqSize(vlistInqVarZaxis(vlistID2, varID));
	  vardata2[varID]  = (double *) malloc(nlev*gridsize*sizeof(double));
	  varnmiss2[varID] = (int *) malloc(nlev*sizeof(int));
	}
    }

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      if ( tsID == 0 || lptype == 0 )
	{
	  nrecs2 = streamInqTimestep(streamID2, tsID);
	  if ( nrecs2 == 0 )
	    cdoAbort("Input streams have different number of timesteps!");
	}

      streamDefTimestep(streamID3, tsID);

      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamReadRecord(streamID1, field1.ptr, &field1.nmiss);

	  if ( tsID == 0 || lptype == 0 )
	    {
	      if ( recID == 0 || lptype != 2 )
		{
		  streamInqRecord(streamID2, &varID, &levelID);
		  streamReadRecord(streamID2, field2.ptr, &field2.nmiss);
		}

	      if ( lptype == 1 )
		{
		  gridsize = gridInqSize(vlistInqVarGrid(vlistID2, varID));
		  offset   = gridsize*levelID;
		  memcpy(vardata2[varID]+offset, field2.ptr, gridsize*sizeof(double));
		  varnmiss2[varID][levelID] = field2.nmiss;
		}
	    }
	  else if ( lptype == 1 )
	    {
	      gridsize = gridInqSize(vlistInqVarGrid(vlistID2, varID));
	      offset   = gridsize*levelID;
	      memcpy(field2.ptr, vardata2[varID]+offset, gridsize*sizeof(double));
	      field2.nmiss = varnmiss2[varID][levelID];
	    }

	  field1.grid    = vlistInqVarGrid(vlistID1, varID);
	  field1.missval = vlistInqVarMissval(vlistID1, varID);

	  if ( recID == 0 || lptype != 2 )
	    {
	      field2.grid    = vlistInqVarGrid(vlistID2, varID);
	      field2.missval = vlistInqVarMissval(vlistID2, varID);
	    }

	  farfun(&field1, field2, function[operatorID]);

	  streamDefRecord(streamID3, varID, levelID);
	  streamWriteRecord(streamID3, field1.ptr, field1.nmiss);
	}
      tsID++;
    }

  streamClose(streamID3);
  streamClose(streamID2);
  streamClose(streamID1);

  if ( vardata2 )
    {
      for ( varID = 0; varID < nvars; varID++ )
	{
	  free(vardata2[varID]);
	  free(varnmiss2[varID]);
	}

      free(vardata2);
      free(varnmiss2);
    }

  if ( field1.ptr ) free(field1.ptr);
  if ( field2.ptr ) free(field2.ptr);

  cdoFinish();

  return (0);
}
