/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Change
@Title     = Change
@Section   = Manipulating the header/field
@Arguments = ifile ofile
@Operators = chcode chvar

@EndModule


@BeginOperator_chcode

@Title     = Change code
@Parameter = ocode ncode ...

@BeginDesciption
Changes some user given codes to new user given values.
@EndDesciption

@BeginParameter
@Item = ocode,ncode,...
INTEGER  Pairs of old and new code
@EndParameter

@EndOperator


@BeginOperator_chname

@Title     = Change variable name
@Parameter = ovar nvar ...

@BeginDesciption
Changes some user given variable names to new user given names.
@EndDesciption

@BeginParameter
@Item = ovar,nvar,...
STRING  Pairs of old and new variable name
@EndParameter

@EndOperator

@EndDoc
*/

#define  MAXARG     16384
#define  NOPERATORS     2

void *Change(void *argument)
{
  static char func[] = "Change";
  const int noperators = NOPERATORS;
  char *operators[NOPERATORS];
  int operatorID;
  char *ostring[NOPERATORS];
  int function[NOPERATORS];
  int streamID1, streamID2 = CDI_UNDEFID;
  int nrecs, nvars;
  int tsID1, recID, varID, levelID;
  int vlistID1, vlistID2;
  int taxisID1;
  int chcodes[MAXARG], nch = 0;
  char *chvars[MAXARG];
  char varname[128];
  int code, i;
  int nmiss;
  int gridsize;
  double *array = NULL;

  i = 0;
  operators[i] = "chcode";   ostring[i] = "pairs of old and new code";           function[i++] = func_code;
  operators[i] = "chvar";    ostring[i] = "pairs of old and new variable name";  function[i++] = func_name;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  operatorInputArg(ostring[operatorID]);

  nch = operatorArgc();
  if ( nch%2 ) cdoAbort("Odd number of input arguments!");

  if ( function[operatorID] == func_code )
    {
      for ( i = 0; i < nch; i++ )
	chcodes[i] = atoi(operatorArgv()[i]);
    }
  else if ( function[operatorID] == func_name )
    {
      for ( i = 0; i < nch; i++ )
	chvars[i] = operatorArgv()[i];
    }

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  taxisID1 = vlistInqTaxis(vlistID1);

  vlistID2 = vlistDuplicate(vlistID1);

  nvars = vlistNvars(vlistID2);
  for ( varID = 0; varID < nvars; varID++ )
    {
      if ( function[operatorID] == func_code )
	{
	  code = vlistInqVarCode(vlistID2, varID);
	  for ( i = 0; i < nch; i += 2 )
	    if ( code == chcodes[i] )
	      vlistDefVarCode(vlistID2, varID, chcodes[i+1]);
	}
      else if ( function[operatorID] == func_name )
	{
	  vlistInqVarName(vlistID2, varID, varname);
	  for ( i = 0; i < nch; i += 2 )
	    if ( strcmp(varname, chvars[i]) == 0 )
	      vlistDefVarName(vlistID2, varID, chvars[i+1]);
	}
    }

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  gridsize = vlistGridsizeMax(vlistID2);
  array = (double *) malloc(gridsize*sizeof(double));

  tsID1 = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID1)) )
    {
      streamDefTimestep(streamID2, tsID1);
	       
      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamDefRecord(streamID2,  varID,  levelID);
	  
	  streamReadRecord(streamID1, array, &nmiss);
	  streamWriteRecord(streamID2, array, nmiss);
	}
      tsID1++;
    }

  streamClose(streamID1);
  streamClose(streamID2);

  if ( array ) free(array);

  cdoFinish();

  return (0);
}
