/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Compc
@Title     = Comparison with a constant
@Section   = Comparisons
@Arguments = ifile ofile
@Operators = eqc nec lec ltc gec gtc

@EndModule

@BeginOperator_eqc

@Title     = Equal constant
@Parameter = c

@BeginDesciption
@IfMan
          /   1   if i(t,x) EQ c     AND  i(t,x),c NE miss
o(t,x) = <    0   if i(t,x) NE c     AND  i(t,x),c NE miss
          \  miss if i(t,x) EQ miss  OR   c EQ miss
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
  1   & \mbox{if} \;\; i(t,x) = \mbox{c}    & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}      \\
  0   & \mbox{if} \;\; i(t,x) \neq \mbox{\sl c} & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
 \mbox{miss} & \mbox{if} \;\; i(t,x) = \mbox{miss} & \vee   \;\; \mbox{\sl c} = \mbox{miss}          \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = c
FLOAT  Constant
@EndParameter

@EndOperator


@BeginOperator_nec

@Title     = Not equal constant
@Parameter = c

@BeginDesciption
@IfMan
          /   1   if i(t,x) NE c     AND  i(t,x),c NE miss
o(t,x) = <    0   if i(t,x) EQ c     AND  i(t,x),c NE miss
          \  miss if i(t,x) EQ miss  OR   c EQ miss
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
  1   & \mbox{if} \;\; i(t,x) \neq \mbox{\sl c} & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
  0   & \mbox{if} \;\; i(t,x) = \mbox{\sl c}    & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
 \mbox{miss} & \mbox{if} \;\; i(t,x) = \mbox{miss} & \vee   \;\; \mbox{\sl c} = \mbox{miss}          \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = c
FLOAT  Constant
@EndParameter

@EndOperator


@BeginOperator_lec

@Title     = Less equal constant
@Parameter = c

@BeginDesciption
@IfMan
          /   1   if i(t,x) LE c     AND  i(t,x),c NE miss
o(t,x) = <    0   if i(t,x) GT c     AND  i(t,x),c NE miss
          \  miss if i(t,x) EQ miss  OR   c EQ miss
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
  1   & \mbox{if} \;\; i(t,x) \leq \mbox{\sl c} & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
  0   & \mbox{if} \;\; i(t,x) > \mbox{\sl c}    & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
 \mbox{miss} & \mbox{if} \;\; i(t,x) = \mbox{miss} & \vee   \;\; \mbox{\sl c} = \mbox{miss}          \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = c
FLOAT  Constant
@EndParameter

@EndOperator


@BeginOperator_ltc

@Title     = Less then constant
@Parameter = c

@BeginDesciption
@IfMan
          /   1   if i(t,x) LT c     AND  i(t,x),c NE miss
o(t,x) = <    0   if i(t,x) GE c     AND  i(t,x),c NE miss
          \  miss if i(t,x) EQ miss  OR   c EQ miss
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
  1   & \mbox{if} \;\; i(t,x) < \mbox{\sl c}    & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
  0   & \mbox{if} \;\; i(t,x) \geq \mbox{\sl c} & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
 \mbox{miss} & \mbox{if} \;\; i(t,x) = \mbox{miss} & \vee   \;\; \mbox{\sl c} = \mbox{miss}          \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = c
FLOAT  Constant
@EndParameter

@EndOperator


@BeginOperator_gec

@Title     = Greater equal constant
@Parameter = c

@BeginDesciption
@IfMan
          /   1   if i(t,x) GE c     AND  i(t,x),c NE miss
o(t,x) = <    0   if i(t,x) LT c     AND  i(t,x),c NE miss
          \  miss if i(t,x) EQ miss  OR   c EQ miss
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
  1   & \mbox{if} \;\; i(t,x) \geq \mbox{\sl c} & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
  0   & \mbox{if} \;\; i(t,x) < \mbox{\sl c}    & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
 \mbox{miss} & \mbox{if} \;\; i(t,x) = \mbox{miss} & \vee   \;\; \mbox{\sl c} = \mbox{miss}          \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = c
FLOAT  Constant
@EndParameter

@EndOperator


@BeginOperator_gtc

@Title     = Greater then constant
@Parameter = c

@BeginDesciption
@IfMan
          /   1   if i(t,x) GT c     AND  i(t,x),c NE miss
o(t,x) = <    0   if i(t,x) LE c     AND  i(t,x),c NE miss
          \  miss if i(t,x) EQ miss  OR   c EQ miss
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
  1   & \mbox{if} \;\; i(t,x) > \mbox{\sl c}    & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
  0   & \mbox{if} \;\; i(t,x) \leq \mbox{\sl c} & \wedge \;\; i(t,x), \mbox{\sl c} \neq \mbox{miss}  \\
 \mbox{miss} & \mbox{if} \;\; i(t,x) = \mbox{miss} & \vee   \;\; \mbox{\sl c} = \mbox{miss}          \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = c
FLOAT  Constant
@EndParameter

@EndOperator

@EndDoc
*/

void *Compc(void *argument)
{
  static char func[] = "Compc";
  const char *operators[] = {"eqc", "nec", "lec", "ltc", "gec", "gtc"};
  enum                     {f_eqc, f_nec, f_lec, f_ltc, f_gec, f_gtc};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int streamID1, streamID2;
  int gridsize;
  int nrecs, recID;
  int tsID;
  int varID, levelID;
  int vlistID1, vlistID2;
  int nmiss, nmiss2;
  int i;
  double missval;
  double rc;
  double *array1, *array2;

  cdoInitialize(argument);

  operatorInputArg("constant value");
  rc = atof(operatorArgv()[0]);

  operatorID = cdoOperatorID(noperators, operators);

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);

  nospec(vlistID1);

  gridsize = vlistGridsizeMax(vlistID1);

  array1 = (double *) malloc(gridsize*sizeof(double));
  array2 = (double *) malloc(gridsize*sizeof(double));

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      streamDefTimestep(streamID2, tsID);

      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamReadRecord(streamID1, array1, &nmiss);

	  missval  = vlistInqVarMissval(vlistID1, varID);
	  gridsize = gridInqSize(vlistInqVarGrid(vlistID1, varID));

	  switch (operatorID)
	    {
	    case f_eqc:
	      for ( i = 0; i < gridsize; i++ )
		array2[i] = array1[i] == missval || rc == missval ? missval : array1[i] == rc;
	      break;
	    case f_nec:
	      for ( i = 0; i < gridsize; i++ )
		array2[i] = array1[i] == missval || rc == missval ? missval : array1[i] != rc;
	      break;
	    case f_lec:
	      for ( i = 0; i < gridsize; i++ )
		array2[i] = array1[i] == missval || rc == missval ? missval : array1[i] <= rc;
	      break;
	    case f_ltc:
	      for ( i = 0; i < gridsize; i++ )
		array2[i] = array1[i] == missval || rc == missval ? missval : array1[i] < rc;
	      break;
	    case f_gec:
	      for ( i = 0; i < gridsize; i++ )
		array2[i] = array1[i] == missval || rc == missval ? missval : array1[i] >= rc;
	      break;
	    case f_gtc:
	      for ( i = 0; i < gridsize; i++ )
		array2[i] = array1[i] == missval || rc == missval ? missval : array1[i] > rc;
	      break;
	    default:
	      cdoAbort("operator not implemented!");
	    }

	  nmiss2 = 0;
	  for ( i = 0; i < gridsize; i++ )
	    if ( array2[i] == missval ) nmiss2++;

	  streamDefRecord(streamID2, varID, levelID);
	  streamWriteRecord(streamID2, array2, nmiss2);
	}
      tsID++;
    }

  streamClose(streamID2);
  streamClose(streamID1);

  if ( array2 ) free(array2);
  if ( array1 ) free(array1);

  cdoFinish();

  return (0);
}
