/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Cond
@Title     = 
@Section   = Conditions
@Arguments = ifile1 ifile2 ofile
@Operators = ifthen ifnotthen

@BeginDesciption
A value not equal to zero is treated as "true", zero is treated a "false".
@EndDesciption

@EndModule


@BeginOperator_ifthen

@Title     = If then

@BeginDesciption
@IfMan
         / i_2(t,x) if i_1(t,x) NE 0  AND  i_1(t,x) NE miss
o(t,x) =
         \ miss     if i_1(t,x) EQ 0  OR   i_1(t,x) EQ miss
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
   i_2(t,x) & \mbox{if} \;\;  i_1(t,x) \neq 0 & \wedge \;\; i_1(t,x) \neq \mbox{miss} \\
\mbox{miss} & \mbox{if} \;\;  i_1(t,x)    = 0 & \vee   \;\; i_1(t,x)    = \mbox{miss} \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_ifnotthen

@Title     = If not then

@BeginDesciption
@IfMan
         / i_2(t,x) if i_1(t,x) EQ 0  AND  i_1(t,x) NE miss
o(t,x) = 
         \ miss     if i_1(t,x) NE 0  OR   i_1(t,x) EQ miss
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
   i_2(t,x) & \mbox{if} \;\;  i_1(t,x)    = 0 & \wedge \;\; i_1(t,x) \neq \mbox{miss} \\
\mbox{miss} & \mbox{if} \;\;  i_1(t,x) \neq 0 & \vee   \;\; i_1(t,x)    = \mbox{miss} \\
\end{array}   \right.
@EndMath
@EndifDoc

@EndDesciption

@EndOperator

@EndDoc
*/

void *Cond(void *argument)
{
  static char func[] = "Cond";
  const char *operators[] = {"ifthen", "ifnotthen"};
  enum                     {f_ifthen, f_ifnotthen};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int streamID1, streamID2, streamID3;
  int gridsize;
  int nrecs, nrecs2, nvars = 0, nlev, recID;
  int tsID;
  int varID, levelID;
  int offset;
  int vlistID1, vlistID2, vlistID3;
  int nmiss, nmiss2, nmiss3;
  int i;
  int lptype = 0;
  double missval1, missval2;
  double *array1, *array2, *array3;
  int **varnmiss2 = NULL;
  double **vardata2 = NULL;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, operators);

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));
  streamID2 = streamOpenRead(cdoStreamName(1));
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = streamInqVlist(streamID2);
  vlistID3 = vlistDuplicate(vlistID1);

  if ( vlistNrecs(vlistID2) == 1 && vlistNrecs(vlistID1) != 1 )
    lptype = 2;
  else
    vlistCompare(vlistID1, vlistID2, func_sft);

  nospec(vlistID1);

  streamID3 = streamOpenWrite(cdoStreamName(2), cdoFiletype());
  if ( streamID3 < 0 ) cdiError(streamID3, "Open failed on %s", cdoStreamName(2));

  streamDefVlist(streamID3, vlistID3);

  gridsize = vlistGridsizeMax(vlistID1);

  array1 = (double *) malloc(gridsize*sizeof(double));
  array2 = (double *) malloc(gridsize*sizeof(double));
  array3 = (double *) malloc(gridsize*sizeof(double));

  if ( cdoVerbose )
    cdoPrint("Number of timesteps: file1 %d, file2 %d", vlistNtsteps(vlistID1), vlistNtsteps(vlistID2));

  if ( vlistNtsteps(vlistID2) == 1 && vlistNtsteps(vlistID1) != 1 && lptype == 0 )
    {
      lptype = 1;
      nvars  = vlistNvars(vlistID2);
      vardata2  = (double **) malloc(nvars*sizeof(double *));
      varnmiss2 = (int **) malloc(nvars*sizeof(int *));
      for ( varID = 0; varID < nvars; varID++ )
	{
	  gridsize = gridInqSize(vlistInqVarGrid(vlistID2, varID));
	  nlev     = zaxisInqSize(vlistInqVarZaxis(vlistID2, varID));
	  vardata2[varID]  = (double *) malloc(nlev*gridsize*sizeof(double));
	  varnmiss2[varID] = (int *) malloc(nlev*sizeof(int));
	}
    }

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      if ( tsID == 0 || lptype == 0 )
	{
	  nrecs2 = streamInqTimestep(streamID2, tsID);
	  if ( nrecs2 == 0 )
	    cdoAbort("Input streams have different number of timesteps!");
	}

      streamDefTimestep(streamID3, tsID);

      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamReadRecord(streamID1, array1, &nmiss);

	  if ( tsID == 0 || lptype == 0 )
	    {
	      if ( recID == 0 || lptype != 2 )
		{
		  streamInqRecord(streamID2, &varID, &levelID);
		  streamReadRecord(streamID2, array2, &nmiss2);
		}

	      if ( lptype == 1 )
		{
		  gridsize = gridInqSize(vlistInqVarGrid(vlistID2, varID));
		  offset   = gridsize*levelID;
		  memcpy(vardata2[varID]+offset, array2, gridsize*sizeof(double));
		  varnmiss2[varID][levelID] = nmiss2;
		}
	    }
	  else if ( lptype == 1 )
	    {
	      gridsize = gridInqSize(vlistInqVarGrid(vlistID2, varID));
	      offset   = gridsize*levelID;
	      memcpy(array2, vardata2[varID]+offset, gridsize*sizeof(double));
	      nmiss2 = varnmiss2[varID][levelID];
	    }

	  gridsize = gridInqSize(vlistInqVarGrid(vlistID1, varID));
	  missval1 = vlistInqVarMissval(vlistID1, varID);

	  if ( recID == 0 || lptype != 2 )
	    {
	      /* gridsize = gridInqSize(vlistInqVarGrid(vlistID2, varID)); */
	      missval2  = vlistInqVarMissval(vlistID2, varID);
	    }

	  switch (operatorID)
	    {
	    case f_ifthen:
	      for ( i = 0; i < gridsize; i++ )
		array3[i] = array1[i] != missval1 && array1[i] ? array2[i] : missval1;
	      break;
	    case f_ifnotthen:
	      for ( i = 0; i < gridsize; i++ )
		array3[i] = array1[i] != missval1 && !array1[i] ? array2[i] : missval1;
	      break;
	    default:
	      cdoAbort("operator not implemented!");
	    }

	  nmiss3 = 0;
	  for ( i = 0; i < gridsize; i++ )
	    if ( array3[i] == missval1 ) nmiss3++;

	  streamDefRecord(streamID3, varID, levelID);
	  streamWriteRecord(streamID3, array3, nmiss3);
	}
      tsID++;
    }

  streamClose(streamID3);
  streamClose(streamID2);
  streamClose(streamID1);

  if ( vardata2 )
    {
      for ( varID = 0; varID < nvars; varID++ )
	{
	  free(vardata2[varID]);
	  free(varnmiss2[varID]);
	}

      free(vardata2);
      free(varnmiss2);
    }

  if ( array3 ) free(array3);
  if ( array2 ) free(array2);
  if ( array1 ) free(array1);

  cdoFinish();

  return (0);
}
