/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"
#include "functs.h"

/*
@BeginDoc

@BeginModule

@Name      = Fldstat
@Title     = Field statistic
%@Section   = Statistical description of the fields
@Section   = Statistical description of the data
@Arguments = ifile ofile
@Operators = fldmin fldmax fldsum fldmean fldavg fldstd fldvar

@BeginDesciption
In this program there is the different notion of "mean" and "average"
to distinguish two different kinds of treatment of missing values:
While computing the mean, only the not missing values are considered
to belong to the sample with the side effect of a probably reduced sample
size. Computing the average is just adding the sample members and divide
the result by the sample size. For example, the mean of 1, 2, miss and 3
is (1+2+3)/3 = 2, whereas the average is (1+2+miss+3)/4 = miss/4 = miss.
If there are no missing values in the sample, the average and the mean are
identical.
In this chapter the abbreviations as in the following table are used:

\vspace{3mm}

\fbox{\parbox{15cm}{
\begin{eqnarray*}
\begin{array}{l}
\makebox[3cm][l]{{\bf mean} resp. {\bf avg}} \\
\end{array}
 &  &
 n^{-1} \sum_{i=1}^{n} x_i 
\\
\begin{array}{l}
\makebox[3cm][l]{{\bf mean} resp. {\bf avg}} \\
\mbox{weighted by} \\
\{w_i, i=1, ..., n\}  \\
\end{array}
 & &
  \left ( \sum_{j=1}^{n} w_j \right )^{-1} \sum\limits_{i=1}^{n} w_i \, x_i \\
\\
\begin{array}{l}
%\makebox[3cm][l]{Standard deviation} \\
\makebox[3cm][l]{Variance} \\
\makebox[3cm][l]{{\bf var}} \\
\end{array}
 &  &
 n^{-1} \sum_{i=1}^{n} (x_i - \overline{x})^2
\\
\begin{array}{l}
\makebox[3cm][l]{{\bf var} weighted by} \\
\{w_i, i=1, ..., n\}  \\
\end{array}
 & &
  \left ( \sum_{j=1}^{n} w_j \right )^{-1} \sum\limits_{i=1}^{n} w_i \, 
  \left ( x_i - \left ( \sum_{j=1}^{n} w_j \right )^{-1} \sum\limits_{j=1}^{n} w_j \, x_j \right)^2 \\
\end{eqnarray*}
}}

@EndDesciption

@EndModule


@BeginOperator_fldmin

@Title     = Field minimum

@BeginDesciption
@IfMan
o(t,1) = min{i(t',x'), t'=t}
@EndifMan
@IfDoc
@BeginMath
o(t,1) = \mbox{\bf min}\{i(t',x'), t' = t\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator

@BeginOperator_fldmax

@Title     = Field maximum

@BeginDesciption
@IfMan
o(t,1) = max{i(t',x'), t'=t}
@EndifMan
@IfDoc
@BeginMath
o(t,1) = \mbox{\bf max}\{i(t',x'), t' = t\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator

@BeginOperator_fldsum

@Title     = Field sum

@BeginDesciption
@IfMan
o(t,1) = sum{i(t,x)}
@EndifMan
@IfDoc
@BeginMath
o(t,1) = \sum\limits_x i(t,x)
@EndMath
@EndifDoc
@EndDesciption

@EndOperator

@BeginOperator_fldmean

@Title     = Field mean

@BeginDesciption
@IfMan
o(t,1) = mean{i(t',x'), t'=t}
@EndifMan
@IfDoc
@BeginMath
o(t,1) = \mbox{\bf mean}\{i(t',x'), t' = t\}
@EndMath
@EndifDoc
weighted by area weights obtained by the input field.
@EndDesciption

@EndOperator

@BeginOperator_fldavg

@Title     = Field average

@BeginDesciption
@IfMan
o(t,1) = avg{i(t',x'), t'=t}
@EndifMan
@IfDoc
@BeginMath
o(t,1) = \mbox{\bf avg}\{i(t',x'), t' = t\}
@EndMath
@EndifDoc
weighted by area weights obtained by the input field.
@EndDesciption

@EndOperator

@BeginOperator_fldvar

@Title     = Field variance

@BeginDesciption
@IfMan
o(t,1) = var{i(t',x'), t'=t}
@EndifMan
@IfDoc
@BeginMath
o(t,1) = \mbox{\bf var}\{i(t',x'), t' = t\}
@EndMath
@EndifDoc
weighted by area weights obtained by the input field.
@EndDesciption

@EndOperator

@BeginOperator_fldstd

@Title     = Field standard deviation

@BeginDesciption
@IfMan
o(t,1) = sqrt{var{i(t',x'), t'=t}}
@EndifMan
@IfDoc
@BeginMath
o(t,1) = \sqrt{\mbox{\bf var}\{i(t',x'), t' = t\}}
@EndMath
@EndifDoc
weighted by area weights obtained by the input field.
@EndDesciption

@EndOperator

@EndDoc
*/


#define  NOPERATORS  7

void *Fldstat(void *argument)
{
  static char func[] = "Fldstat";
  const int noperators = NOPERATORS;
  char *operators[NOPERATORS];
  int operatorID;
  int function[NOPERATORS];
  int streamID1, streamID2;
  int vlistID1, vlistID2;
  int gridID1, gridID2, lastgrid = -1;
  int index, ngrids;
  int recID, nrecs;
  int i;
  int tsID, varID, levelID;
  int lim;
  int needWeights = FALSE;
  double slon, slat;
  double sglval;
  FIELD field1, field2;

  i = 0;
  operators[i] = "fldmin";   function[i++] = func_min;
  operators[i] = "fldmax";   function[i++] = func_max;
  operators[i] = "fldsum";   function[i++] = func_sum;
  operators[i] = "fldmean";  function[i++] = func_mean;
  operators[i] = "fldavg";   function[i++] = func_avg;
  operators[i] = "fldvar";   function[i++] = func_var;
  operators[i] = "fldstd";   function[i++] = func_std;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  if ( function[operatorID] == func_mean ||
       function[operatorID] == func_avg  ||
       function[operatorID] == func_var  ||
       function[operatorID] == func_std  )
    needWeights = TRUE;

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);

  vlistID2 = vlistDuplicate(vlistID1);

  slon = 0;
  slat = 0;
  gridID2 = gridNew(GRID_LONLAT, 1);
  gridDefXsize(gridID2, 1);
  gridDefYsize(gridID2, 1);
  gridDefXvals(gridID2, &slon);
  gridDefYvals(gridID2, &slat);

  ngrids = vlistNgrids(vlistID1);
  index = 0;
  gridID1 = vlistGrid(vlistID1, index);
  
  if ( needWeights &&
       gridInqType(gridID1) != GRID_LONLAT &&
       gridInqType(gridID1) != GRID_GAUSSIAN )
    cdoAbort("Unsupported gridtype: %s", gridNamePtr(gridInqType(gridID1)));

  vlistChangeGridIndex(vlistID2, index, gridID2);
  if ( ngrids > 1 ) cdoAbort("Too many different grids!");

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  lim = vlistGridsizeMax(vlistID1);
  field1.ptr    = (double *) malloc(lim*sizeof(double));
  field1.weight = NULL;
  if ( needWeights )
    field1.weight = (double *) malloc(lim*sizeof(double));

  field2.ptr  = &sglval;
  field2.grid = gridID2;

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      streamDefTimestep(streamID2, tsID);

      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamReadRecord(streamID1, field1.ptr, &field1.nmiss);

	  field1.grid    = vlistInqVarGrid(vlistID1, varID);
	  if ( field1.grid != lastgrid )
	    {
	      lastgrid = field1.grid;
	      if ( needWeights ) gridArea(field1.grid, field1.weight);
	    }
	  field1.missval = vlistInqVarMissval(vlistID1, varID);
	  field2.missval = vlistInqVarMissval(vlistID1, varID);

	  fldfun(field1, &field2, function[operatorID]);

	  streamDefRecord(streamID2, varID,  levelID);
	  streamWriteRecord(streamID2, &sglval, field2.nmiss);
	}
      tsID++;
    }

  streamClose(streamID2);
  streamClose(streamID1);

  if ( field1.ptr )    free(field1.ptr);
  if ( field1.weight ) free(field1.weight);

  cdoFinish();

  return (0);
}
