/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#if  defined  (HAVE_CONFIG_H)
#  include "config.h"
#endif

#include <string.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "specspace.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Gradsdes
@Title     = GrADS data descriptor file
@Section   = Other
@Arguments = ifile
@Operators = gradsdes

@EndModule


@BeginOperator_gradsdes

@Title     = GrADS data descriptor file

@BeginDesciption
Creates a GrADS data descriptor file. Supported file formats
are GRIB, SERVICE and EXTRA. For GRIB files the GrADS index 
file is also generated. For SERVICE and EXTRA files the grid
musst be specified with the CDO option '-g <grid>'.
This operator takes ifile in order to create filenames for
the descriptor (ifile.ctl) and the index (ifile.gmp) file.
@EndDesciption

@EndOperator

@EndDoc
*/

#define  CDI_BIGENDIAN            0   /* Data type BIGENDIAN     */
#define  CDI_LITTLEENDIAN         1   /* Data type LITTLEENDIAN  */


void *Gradsdes(void *argument)
{
  static char func[] = "Gradsdes";
  int streamID = 0;
  int gridID = -1, zaxisID;
  int zaxisIDmax = -1, nlevmax;
  int gridtype = -1;
  int nvars, ngrids, nzaxis, nlev;
  int nvarsout;
  int index;
  int vlistID, tsID, varID;
  int recID, levelID;
  int filetype, byteorder;
  int taxisID, ntsteps, nrecs;
  int vdate, vtime;
  int ltype, code;
  const char *datfile;
  char ctlfile[1024];
  size_t len;
  char varname[256], varlongname[256], varunits[256];
  FILE *gdp;
  int yrev = FALSE;
  int zrev = FALSE;
  int i, j;
  int xsize = 0, ysize = 0;
  int prec, res;
  int lplev = FALSE;
  int xyheader = 0;
  int nrecords = 0;
  int bigendian = FALSE, littleendian = FALSE;
  int sequential = FALSE;
  double *levels, level0, levinc = 0;
  double missval = 0;
  char Time[30], Incr[10] = {"1mn"}, *IncrKey[] = {"mn","hr","dy","mo","yr"};
  int imn, imns = 0, ihh, ihhs = 0, iyy, iyys = 0, imm, imms = 0, idd, idds = 0;
  int idmn, idhh, idmm, idyy, iddd;
  int dt=1, iik=0;
  int gridsize = 0;
  int nmiss;
  int maxrecs = 0;
  int *intnum = NULL;
  int *vars = NULL;
  float *fltnum = NULL;
  double *array = NULL;
  double xfirst, yfirst, xinc, yinc;
  double *xvals, *yvals;
  char *cmons[]={"jan","feb","mar","apr","may","jun",
		 "jul","aug","sep","oct","nov","dec"};
      
  cdoInitialize(argument);

  if ( cdoStreamName(0)[0] == '-' )
    cdoAbort("This operator does not work with pipes!");

  streamID = streamOpenRead(cdoStreamName(0));
  if ( streamID < 0 ) cdiError(streamID, "Open failed on %s", cdoStreamName(0));

  vlistID = streamInqVlist(streamID);

  nvars  = vlistNvars(vlistID);
  ngrids = vlistNgrids(vlistID);
  nzaxis = vlistNzaxis(vlistID);

  filetype  = streamInqFiletype(streamID);
  byteorder = streamInqByteorder(streamID);

  if ( filetype != FILETYPE_SRV &&
       filetype != FILETYPE_EXT &&
       filetype != FILETYPE_GRB )
    cdoAbort("unsupported file format!");

  /* find the first lonlat or gaussian grid */

  for ( index = 0; index < ngrids; index++ )
    {
      gridID = vlistGrid(vlistID, index);
      gridtype = gridInqType(gridID);
      if ( gridtype == GRID_LONLAT || gridtype == GRID_GAUSSIAN ) break;
    }

  if ( index == ngrids )
    cdoAbort("No lonlat or gaussian grid found!");

  /* select all variables with used gridID */

  vars = (int *) malloc(nvars*sizeof(int));
  nvarsout = 0;
  for ( varID = 0; varID < nvars; varID++ )
    {
      if ( vlistInqVarGrid(vlistID, varID) == gridID )
	{
	  nvarsout++;
	  vars[varID] = TRUE;
	}
      else
	{
	  vlistInqVarName(vlistID, varID, varname);
	  cdoPrint("%s grid unsupported, skipped variable %s",
		   gridNamePtr(gridInqType(vlistInqVarGrid(vlistID, varID))), varname);
	  vars[varID] = FALSE;
	}
    }

  if ( filetype != FILETYPE_GRB && nvars != nvarsout )
    cdoAbort("Too many different grids!");    

  if ( filetype == FILETYPE_SRV )
    {
      xyheader = 40;
      sequential = TRUE;
      if ( byteorder == CDI_BIGENDIAN )    bigendian = TRUE;
      if ( byteorder == CDI_LITTLEENDIAN ) littleendian = TRUE;
    }

  if ( filetype == FILETYPE_EXT )
    {
      xyheader = 24;
      sequential = TRUE;
      if ( byteorder == CDI_BIGENDIAN )    bigendian = TRUE;
      if ( byteorder == CDI_LITTLEENDIAN ) littleendian = TRUE;
    }

  strcpy(ctlfile, cdoStreamName(0));
  len = strlen(ctlfile);
  if ( len > 4 )
    {
      if ( filetype == FILETYPE_SRV )
	if ( strcmp(&ctlfile[len-4], ".srv") == 0 ) ctlfile[len-4] = 0;
      if ( filetype == FILETYPE_EXT )
	if ( strcmp(&ctlfile[len-4], ".ext") == 0 ) ctlfile[len-4] = 0;
      if ( filetype == FILETYPE_GRB )
	if ( strcmp(&ctlfile[len-4], ".grb") == 0 ) ctlfile[len-4] = 0;
    }

  strcat(ctlfile, ".ctl");
  gdp = fopen(ctlfile, "w");
  if ( gdp == NULL ) cdoAbort("Open failed on %s", gdp);

  fprintf(gdp, "* Generated automatically by cdo version %s\n", VERSION);
  fprintf(gdp, "*\n");

  /* DSET */

  datfile = cdoStreamName(0);
  if ( datfile[0] == '/' )
    fprintf(gdp, "DSET  %s\n", datfile);
  else
    {
      datfile = strrchr(datfile, '/');
      if ( datfile == 0 ) datfile = cdoStreamName(0);
      else                datfile++;	  
      fprintf(gdp, "DSET  ^%s\n", datfile);
    }

  /* DTYPE */

  if ( filetype == FILETYPE_GRB )
    {
      fprintf(gdp, "DTYPE  GRIB\n");

      len = strlen(ctlfile);
      strcpy(&ctlfile[len-4], ".gmp");
      
      fprintf(gdp, "INDEX  ^%s\n", ctlfile);

      gridsize = vlistGridsizeMax(vlistID);
      array = (double *) malloc(gridsize*sizeof(double));
    }

  /* XYHEADER */
  if ( xyheader )
  fprintf(gdp, "XYHEADER  %d\n", xyheader);

  /* TIME */

  taxisID = vlistInqTaxis(vlistID);
  ntsteps = vlistNtsteps(vlistID);

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID, tsID)) )
    {
      vdate = taxisInqVdate(taxisID);
      vtime = taxisInqVtime(taxisID);

      if ( tsID == 0 )
	{
	  ihhs =  vtime / 100;
	  imns =  vtime - ihhs*100;
	  iyys =  vdate / 10000;
	  imms = (vdate - iyys*10000) / 100;
	  idds =  vdate - iyys*10000 - imms*100;
     
	  if ( imms < 1 || imms > 12 )  imms=1;
	  sprintf (Time, "%02d:%02dZ%02d%s%04d",
		   ihhs, imns, idds, cmons[imms-1], iyys);
	}

      if ( tsID == 1 )
	{
	  ihh =  vtime / 100;
	  imn =  vtime - ihh*100;
	  iyy =  vdate / 10000;
	  imm = (vdate - iyy*10000) / 100;
	  idd =  vdate - iyy*10000 - imm*100;

	  idmn = imn - imns;
	  idhh = ihh - ihhs;
	  iddd = idd - idds;
	  idmm = imm - imms;
	  idyy = iyy - iyys;

	  if ( idmn != 0 )
	    {
	      dt = idmn + (idhh + (iddd + (idmm*30 + idyy*12)*30)*24)*60;
	    }
	  else if ( idhh != 0 )
	    {
	      dt = idhh + (iddd + (idmm + idyy*12)*30)*24;
	      iik = 1;
	    }
	  else if ( iddd != 0 )
	    {
	      dt = iddd + (idmm + idyy*12)*30;
	      iik = 2;
	    }
	  else if ( idmm != 0 )
	    {
	      dt = idmm + idyy*12;
	      iik = 3;
	    }
	  else if ( idyy != 0 )
	    {
	      dt = idyy;
	      iik = 4;
	    }

	  if ( dt <= 0 ) dt = 1;
	  sprintf (Incr, "%d%s", dt, IncrKey[iik]);
	}

      if ( filetype == FILETYPE_GRB )
	{
	  for ( recID = 0; recID < nrecs; recID++ )
	    {
	      streamInqRecord(streamID, &varID, &levelID);
	      if ( vars[varID] == TRUE )
		{
		  streamReadRecord(streamID, array, &nmiss);
	      
		  if ( nrecords >= maxrecs )
		    {
		      maxrecs += 1000;
		      intnum = (int *) realloc(intnum, 3*maxrecs*sizeof(int));
		      fltnum = (float *) realloc(fltnum, 3*maxrecs*sizeof(float));
		    }
		  
		  streamInqGinfo(streamID, &intnum[3*nrecords], &fltnum[3*nrecords]);

		  nrecords += 1;
		}
	    }
	}

      tsID++;
    }

  /* XDEF */

  xsize  = gridInqXsize(gridID);
  xfirst = gridInqXval(gridID, 0);
  xinc   = gridInqXinc(gridID);
  if ( xinc == 0 && gridInqXvals(gridID, NULL) )
    {
      xvals = (double *) malloc(xsize*sizeof(double));
      gridInqXvals(gridID, xvals);
      fprintf(gdp ,"XDEF %d LEVELS ", xsize);
      j = 0;
      for ( i = 0; i < xsize; i++ )
	{
	  fprintf(gdp, "%7.3f ", xvals[i]); 
	  j++;
	  if ( j == 6 )
	    {
	      fprintf(gdp, "\n");
	      j = 0;
	      if ( i != xsize-1 ) fprintf(gdp, "               ");
	    }
	}
      if ( j ) fprintf(gdp, "\n");

      free(xvals);
    }
  else
    {
      if ( xinc == 0 ) xinc = 360.0/xsize;
      fprintf(gdp, "XDEF %d LINEAR %f %f\n", xsize, xfirst, xinc);	  
    }

  /* YDEF */

  ysize  = gridInqYsize(gridID);
  yfirst = gridInqYval(gridID, 0);
  yinc   = gridInqYinc(gridID);
  if ( yinc == 0 && gridInqYvals(gridID, NULL) )
    {
      yvals = (double *) malloc(ysize*sizeof(double));
      gridInqYvals(gridID, yvals);
      fprintf(gdp ,"YDEF %d LEVELS ", ysize);
      j = 0;
      if ( yvals[0] > yvals[ysize-1] )
	{
	  yrev = TRUE;
	  for ( i = ysize-1; i >= 0; i-- )
	    {
	      fprintf(gdp, "%7.3f ", yvals[i]); 
	      j++;
	      if ( j == 6 )
		{
		  fprintf(gdp, "\n");
		  j = 0;
		  if ( i != 0 ) fprintf(gdp, "               ");
		}
	    }
	}
      else
	{
	  for ( i = 0; i < ysize; i++ )
	    {
	      fprintf(gdp, "%7.3f ", yvals[i]); 
	      j++;
	      if ( j == 6 )
		{
		  fprintf(gdp, "\n");
		  j = 0;
		  if ( i != ysize-1 ) fprintf(gdp, "               ");
		}
	    }
	}

      if ( j ) fprintf(gdp, "\n");

      free(yvals);
    }
  else
    {
      if ( yinc == 0 ) yinc = 180.0/ysize;
      fprintf(gdp, "YDEF %d LINEAR %f %f\n", ysize, yfirst, yinc);
    }

  /* ZDEF */

  nlevmax = 0;
  for ( index = 0; index < nzaxis; index++ )
    {
      zaxisID = vlistZaxis(vlistID, index);
      nlev    = zaxisInqSize(zaxisID);
      if ( nlev > nlevmax )
	{
	  nlevmax = nlev;
	  zaxisIDmax = zaxisID;
	}
    }

  levels = (double *) malloc(nlevmax*sizeof(double));
  zaxisInqLevels(zaxisIDmax, levels);
  if ( zaxisInqType(zaxisIDmax) == ZAXIS_PRESSURE ) lplev = TRUE;
  level0 = levels[0];
  if ( nlevmax > 1 )
    {
      if ( levels[0] < levels[1] && zaxisInqType(zaxisIDmax) != ZAXIS_HYBRID )
	zrev = TRUE;

      levinc = levels[1] - levels[0];

      if ( levinc == 1 ) zrev = FALSE;

      for ( i = 1; i < nlevmax; i++ )
	{
	  if ( levinc != (levels[i] - levels[i-1]) )
	    {
	      levinc = 0;
	      break;
	    }
	}
    }

  if ( levinc )
    fprintf(gdp,"ZDEF %d LINEAR %g %g\n", nlevmax, level0, levinc);
  else
    {
      fprintf(gdp, "ZDEF %d LEVELS ", nlevmax);
      j  = 0;
      if ( zrev )
	{
	  for ( i = nlevmax-1; i >=0 ; i-- )
	    {
	      if ( lplev ) fprintf(gdp, "%g ", levels[i]/100);
	      else         fprintf(gdp, "%d ", (int) levels[i]);
	      j++;
	      if ( j == 10 )
		{
		  fprintf(gdp, "\n");
		  j = 0;
		  if ( i != 0 ) fprintf(gdp, "               ");
		}
	    }
	}
      else
	{
	  for ( i = 0; i < nlevmax ; i++ )
	    {
	      if ( lplev ) fprintf(gdp, "%g ", levels[i]/100);
	      else         fprintf(gdp, "%d ", (int) levels[i]);
	      j++;
	      if ( j == 10 )
		{
		  fprintf(gdp, "\n");
		  j = 0;
		  if ( i != (nlevmax-1) ) fprintf(gdp, "               ");
		}
	    }
	}
      if ( j ) fprintf(gdp, "\n");
    }

  free(levels);

  /* TDEF */
  
  fprintf (gdp, "TDEF %d LINEAR %s %s\n", tsID, Time, Incr);

  /* TITLE */

  res = 0;
  if ( gridtype == GRID_GAUSSIAN ) res = nlat2trunc(ysize);

  if ( res )
    fprintf(gdp, "TITLE  %s  T%d grid\n", datfile, res);
  else
    fprintf(gdp, "TITLE  %s  %dx%d grid\n", datfile, xsize, ysize);

  /* OPTIONS */
  
  if ( filetype == FILETYPE_GRB ) zrev = FALSE;

  if ( yrev || zrev || sequential || bigendian || littleendian )
    {
      fprintf(gdp, "OPTIONS");
      if ( yrev )         fprintf(gdp, " yrev");
      if ( zrev )         fprintf(gdp, " zrev");
      if ( sequential )   fprintf(gdp, " sequential");
      if ( bigendian )    fprintf(gdp, " big_endian");
      if ( littleendian ) fprintf(gdp, " little_endian");
      fprintf(gdp, "\n");
    }
    
  /* UNDEF */

  missval = vlistInqVarMissval(vlistID, 0);
  fprintf(gdp, "UNDEF  %g\n", missval);
  
  /* VARS */

  fprintf(gdp, "VARS  %d\n", nvarsout);

  for ( varID = 0; varID < nvars; varID++ )
    {
      if ( vars[varID] == TRUE )
	{
	  if ( filetype == FILETYPE_SRV || filetype == FILETYPE_EXT )
	    {
	      prec = vlistInqVarDatatype(vlistID, varID);
	      if ( prec != DATATYPE_REAL4 )
		cdoAbort("datatype musst be 4 bytes!");
	    }

	  zaxisID = vlistInqVarZaxis(vlistID, varID);
	  ltype   = zaxisInqType(zaxisID);
	  nlev    = zaxisInqSize(zaxisID);
	  vlistInqVarName(vlistID, varID, varname);
	  vlistInqVarLongname(vlistID, varID, varlongname);
	  vlistInqVarUnits(vlistID, varID, varunits);
	  fprintf(gdp, "%-15s", varname);
      
	  if ( nlev == 1 ) nlev = 0;

	  fprintf(gdp, "  %3d", nlev);

	  if ( filetype == FILETYPE_GRB )
	    {
	      code = vlistInqVarCode(vlistID, varID);
	      if      ( ltype == ZAXIS_SURFACE )  ltype = 1;
	      else if ( ltype == ZAXIS_PRESSURE ) ltype = 99;
	      else if ( nlev == 1 )  ltype = 1;
	      else ltype = 99;

	      fprintf(gdp, "  %d,%d,%d", code, ltype, 0);
	    }
	  else
	    fprintf(gdp, "  99");

	  if ( varlongname[0] == 0 )
	    fprintf(gdp, "  %s", varname);
	  else
	    fprintf(gdp, "  %s", varlongname);

	  if ( varunits[0] != 0 )
	    fprintf(gdp, "  [%s]", varunits);
	
	  fprintf(gdp, "\n");      
	}
    }

  fprintf(gdp, "ENDVARS\n");

  /* INDEX file */

  if ( filetype == FILETYPE_GRB )
    {
      struct gaindx {
	int type;                 /* Indexing file type             */
	int hinum;                /* Number of ints in header       */
	int hfnum;                /* Number of floats in header     */
	int intnum;               /* Number of index ints (long)    */
	int fltnum;               /* Number of index floats         */
	int   *hipnt;             /* Pointer to header int values   */
	float *hfpnt;             /* Pointer to header float values */
	int   *intpnt;            /* Pointer to int index values    */
	float *fltpnt;            /* Pointer to float index values  */
      } ;

      struct gaindx indx;
      FILE *ifp;
      int hinum[4];

      ifp = fopen(ctlfile, "w");
      if ( ifp == NULL ) cdoAbort("Open failed on %s", ctlfile);

      indx.type   = 1;  /* GRIB type */
      indx.hinum  = 4;
      indx.hfnum  = 0;
      indx.intnum = 3 * nrecords;
      indx.fltnum = 3 * nrecords;
      indx.hipnt  = NULL;
      indx.hfpnt  = NULL;
      indx.intpnt = NULL;
      indx.fltpnt = NULL;

      fwrite (&indx, sizeof(struct gaindx), 1, ifp);

      hinum[0] = 1;
      hinum[1] = 1;
      hinum[2] = nrecords;
      hinum[3] = 255;
      fwrite(hinum, sizeof(int), 4, ifp);

      fwrite(intnum, sizeof(int), 3*nrecords, ifp);

      fwrite(fltnum, sizeof(float), 3*nrecords, ifp);

      fclose(ifp);
    }

  streamClose(streamID);

  if ( vars ) free(vars);
  if ( array ) free(array);
  if ( intnum ) free(intnum);
  if ( fltnum ) free(fltnum);

  cdoFinish();

  return (0);
}
