/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>
#include <math.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"
#include "interpol.h"

/*
@BeginDoc

@BeginModule

@Name      = Intgrid
@Title     = Grid interpolation
@Section   = Interpolation
@Arguments = ifile ofile
@Operators = interpolate intgrid intpoint

@EndModule


@BeginOperator_interpolate

@Title     = Interpolate
@Parameter = grid

@BeginDesciption
Grid interpolation from PINGO.
@EndDesciption

@BeginParameter
@Item = grid
STRING  Grid description file or name of the target grid
@EndParameter

@EndOperator


@BeginOperator_intgrid

@Title     = Grid interpolation
@Parameter = grid

@BeginDesciption
Linear grid interpolation.
@EndDesciption

@BeginParameter
@Item = grid
STRING  Grid description file or name of the target grid
@EndParameter

@EndOperator


@BeginOperator_intpoint

@Title     = Point interpolation
@Parameter = long lat

@BeginDesciption
Linear point interpolation.
@EndDesciption

@BeginParameter long
@Item = long
FLOAT  Longitude of the point
@Item = lat
FLOAT  Latitude of the point
@EndParameter

@EndOperator

@EndDoc
*/

void *Intgrid(void *argument)
{
  static char func[] = "Intgrid";
  const char *operators[] = {"intgrid", "intpoint", "interpolate"};
  enum                     {f_intgrid, f_intpoint, f_interpolate};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int streamID1, streamID2;
  int nrecs, ngrids;
  int index;
  int tsID, recID, varID, levelID;
  int gridsize;
  int vlistID1, vlistID2;
  int gridID1 = -1, gridID2;
  int nmiss;
  double missval;
  double slon, slat;
  double *array1 = NULL, *array2 = NULL;
  FIELD field1, field2;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, operators);

  if ( operatorID == f_intgrid || operatorID == f_interpolate )
    {
      operatorInputArg("grid description file or name");
      gridID2 = cdoDefineGrid(operatorArgv()[0]);
    }
  else
    {
      operatorInputArg("longitude and latitude");
      if ( operatorArgc() < 2 ) cdoAbort("No enough arguments!");
      slon = atof(operatorArgv()[0]);
      slat = atof(operatorArgv()[1]);
      gridID2 = gridNew(GRID_LONLAT, 1);
      gridDefXsize(gridID2, 1);
      gridDefYsize(gridID2, 1);
      gridDefXvals(gridID2, &slon);
      gridDefYvals(gridID2, &slat);
    }

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);

  ngrids = vlistNgrids(vlistID1);
  for ( index = 0; index < ngrids; index++ )
    {
      gridID1 = vlistGrid(vlistID1, index);

      if ( gridInqType(gridID1) != GRID_LONLAT &&
	   gridInqType(gridID1) != GRID_GAUSSIAN )
	cdoAbort("Interpolation of %s data unsupported!", gridNamePtr(gridInqType(gridID1)) );

      vlistChangeGridIndex(vlistID2, index, gridID2);
    }

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  gridsize = vlistGridsizeMax(vlistID1);
  array1   = (double *) malloc(gridsize*sizeof(double));

  gridsize = gridInqSize(gridID2);
  array2   = (double *) malloc(gridsize*sizeof(double));

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      streamDefTimestep(streamID2, tsID);
	       
      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamReadRecord(streamID1, array1, &nmiss);

	  gridID1 = vlistInqVarGrid(vlistID1, varID);
	  missval = vlistInqVarMissval(vlistID1, varID);

	  field1.grid    = gridID1;
	  field1.nmiss   = nmiss;
	  field1.missval = missval;
	  field1.ptr     = array1;
	  field2.grid    = gridID2;
	  field2.ptr     = array2;
	  field2.nmiss   = 0;

	  if ( operatorID == f_intgrid || operatorID == f_intpoint )
	    intgrid(gridID1, array1, gridID2, array2);
	  else if ( operatorID == f_interpolate )
	    interpolate(&field1, &field2);

	  nmiss = field2.nmiss;

	  streamDefRecord(streamID2, varID, levelID);
	  streamWriteRecord(streamID2, array2, nmiss);
	}
      tsID++;
    }

  streamClose(streamID2);
  streamClose(streamID1);

  if ( array2 ) free(array2);
  if ( array1 ) free(array1);

  cdoFinish();

  return (0);
}
