/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <stdio.h>
#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Output
@Title     = Formatted output
@Section   = Formatted I/O
@Arguments = ifiles
@Operators = output outputint outputsrv outputext

@EndModule


@BeginOperator_output

@Title     = ASCII output

@BeginDesciption
Prints all values to standard output.
@EndDesciption

@EndOperator


@BeginOperator_outputint

@Title     = Integer output

@BeginDesciption
Prints all values rounded to the nearest intergers to standard output.
@EndDesciption

@EndOperator


@BeginOperator_outputsrv

@Title     = SERVICE output

@BeginDesciption
Prints all values to standard output.
Each field with a header of 8 integers (SERVICE likely).
@EndDesciption

@EndOperator


@BeginOperator_outputext

@Title     = EXTRA output

@BeginDesciption
Prints all values to standard output.
Each field with a header of 4 integers (EXTRA likely).
@EndDesciption

@EndOperator

@EndDoc
*/

void *Output(void *argument)
{
  static char func[] = "Output";
  int i;
  const char *operators[] = {"output", "outputint", "outputsrv", "outputext", "outputf", "outputts"};
  enum                     {f_output, f_outputint, f_outputsrv, f_outputext, f_outputf, f_outputts};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int indf;
  int varID, recID;
  int gridsize = 0;
  int gridID, zaxisID, code, vdate, vtime;
  int nrecs;
  int levelID;
  int tsID, taxisID;
  int streamID = 0;
  int vlistID;
  int nmiss, nout;
  int nlon, nlat;
  double level;
  int hour, minute;
  const char *format = NULL;
  double *array = NULL;
  double xdate;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, operators);

  if ( operatorID == f_outputf )
    {
      operatorInputArg("format");
      format = operatorArgv()[0];
    }

  for ( indf = 0; indf < cdoStreamCnt(); indf++ )
    {
      streamID = streamOpenRead(cdoStreamName(indf));
      if ( streamID < 0 ) cdiError(streamID, "Open failed on %s", cdoStreamName(indf));

      vlistID = streamInqVlist(streamID);

      gridsize = vlistGridsizeMax(vlistID);

      array = (double *) malloc(gridsize*sizeof(double));

      tsID = 0;
      taxisID = vlistInqTaxis(vlistID);
      while ( (nrecs = streamInqTimestep(streamID, tsID)) )
	{
	  vdate = taxisInqVdate(taxisID);
	  vtime = taxisInqVtime(taxisID);

	  for ( recID = 0; recID < nrecs; recID++ )
	    {
	      streamInqRecord(streamID, &varID, &levelID);

	      code     = vlistInqVarCode(vlistID, varID);
	      gridID   = vlistInqVarGrid(vlistID, varID);
	      zaxisID  = vlistInqVarZaxis(vlistID, varID);
	      gridsize = gridInqSize(gridID);
	      nlon     = gridInqXsize(gridID);
	      nlat     = gridInqYsize(gridID);
	      level    = zaxisInqLevel(zaxisID, levelID);

	      streamReadRecord(streamID, array, &nmiss);

	      if ( operatorID == f_outputsrv )
		fprintf(stdout, "%4d %8g %8d %4d %8d %8d %d %d\n", code, level, vdate, vtime, nlon, nlat, 0, 0);

	      if ( operatorID == f_outputext )
		fprintf(stdout, "%8d %4d %8g %8d\n", vdate, code, level, gridsize);
		
	      if ( operatorID == f_outputint )
		{
		  nout = 0;
		  for ( i = 0; i < gridsize; i++ )
		    {
		      if ( nout == 8 )
			{
			  nout = 0;
			  fprintf(stdout, "\n");
			}
		      fprintf(stdout, " %8d", (int) array[i]);
		      nout++;
		    }
		  fprintf(stdout, "\n");
		}
	      else if ( operatorID == f_outputf )
		{
		  nout = 0;
		  for ( i = 0; i < gridsize; i++ )
		    {
		      if ( nout == 6 )
			{
			  nout = 0;
			  fprintf(stdout, "\n");
			}
		      fprintf(stdout, format, array[i]);
		      nout++;
		    }
		  fprintf(stdout, "\n");
		}
	      else if ( operatorID == f_outputts )
		{
		  if ( gridsize > 1 )
		    cdoAbort("operator works only with one gridpoint!");

		  hour   = vtime / 100;
		  minute = vtime - hour*100;
		  xdate  = vdate - (vdate/100)*100 + (hour*60 + minute)/1440.;
		  fprintf(stdout, "%12.12g   %12.12g\n", xdate, array[0]);
		}
	      else
		{
		  nout = 0;
		  for ( i = 0; i < gridsize; i++ )
		    {
		      if ( nout == 6 )
			{
			  nout = 0;
			  fprintf(stdout, "\n");
			}
		      fprintf(stdout, " %12.6g", array[i]);
		      nout++;
		    }
		  fprintf(stdout, "\n");
		}
	    }
	  tsID++;
	}
      streamClose(streamID);

      if ( array ) free(array);
    }

  cdoFinish();

  return (0);
}
