/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <stdio.h>
#include <math.h>

#include "cdi.h"
#include "cdo.h"
#include "pstream.h"
#include "dtypes.h"
#include "functs.h"
#include "field.h"
#include "dmemory.h"

/*
@BeginDoc

@BeginModule

@Name      = Runstat
@Title     = 
%@Section   = Statistical description of a time period
@Section   = Statistical description of the data
@Arguments = ifile ofile
@Operators = runmin runmax runsum runmean runavg runstd
@EndModule


@BeginOperator_runmin

@Title     = Running minimum
@Parameter = nts

@BeginDesciption
@IfMan
o(t,x) = min{i(t,x), i(t+1,x), ..., i(t+nts-1,x)}
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \mbox{\bf min}\{i(t,x), i(t+1,x), ..., i(t+nts-1,x)\}
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = nts
INTEGER  Number of timesteps
@EndParameter

@EndOperator


@BeginOperator_runmax

@Title     = Running maximum
@Parameter = nts

@BeginDesciption
@IfMan
o(t,x) = max{i(t,x), i(t+1,x), ..., i(t+nts-1,x)}
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \mbox{\bf max}\{i(t,x), i(t+1,x), ..., i(t+nts-1,x)\}
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = nts
INTEGER  Number of timesteps
@EndParameter

@EndOperator


@BeginOperator_runsum

@Title     = Running sum
@Parameter = nts

@BeginDesciption
@IfMan
o(t,x) = sum{i(t,x), i(t+1,x), ..., i(t+nts-1,x)}
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \sum \{i(t,x), i(t+1,x), ..., i(t+nts-1,x)\}
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = nts
INTEGER  Number of timesteps
@EndParameter

@EndOperator


@BeginOperator_runmean

@Title     = Running mean
@Parameter = nts

@BeginDesciption
@IfMan
o(t,x) = mean{i(t,x), i(t+1,x), ..., i(t+nts-1,x)}
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \mbox{\bf mean}\{i(t,x), i(t+1,x), ..., i(t+nts-1,x)\}
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = nts
INTEGER  Number of timesteps
@EndParameter

@EndOperator


@BeginOperator_runavg

@Title     = Running average
@Parameter = nts

@BeginDesciption
@IfMan
o(t,x) = avg{i(t,x), i(t+1,x), ..., i(t+nts-1,x)}
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \mbox{\bf avg}\{i(t,x), i(t+1,x), ..., i(t+nts-1,x)\}
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = nts
INTEGER  Number of timesteps
@EndParameter

@EndOperator


@BeginOperator_runstd

@Title     = Running standard deviation
@Parameter = nts

@BeginDesciption
@IfMan
o(t,x) = sqrt{var{i(t,x), i(t+1,x), ..., i(t+nts-1,x)}}
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \sqrt{\mbox{\bf var}\{i(t,x), i(t+1,x), ..., i(t+nts-1,x)\}}
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = nts
INTEGER  Number of timesteps
@EndParameter

@EndOperator


@EndDoc
*/

#define  NOPERATORS  6

void *Runstat(void *argument)
{
  static char func[] = "Runstat";
  const int noperators = NOPERATORS;
  char *operators[NOPERATORS];
  int operatorID;
  int function[NOPERATORS];
  int gridsize;
  int varID;
  int recID;
  int gridID;
  int nrecs, nrecords;
  int levelID;
  int tsID;
  int otsID;
  int inp, its, ndates = 0;
  int i;
  int streamID1, streamID2;
  int vlistID1, vlistID2;
  int nmiss;
  int nvars, nlevel;
  int *recVarID, *recLevelID;
  double missval;
  FIELD ***vars1 = NULL, ***vars2 = NULL;

  i = 0;
  operators[i] = "runmin";   function[i] = func_min;  i++;
  operators[i] = "runmax";   function[i] = func_max;  i++;
  operators[i] = "runsum";   function[i] = func_sum;  i++;
  operators[i] = "runmean";  function[i] = func_mean; i++;
  operators[i] = "runavg";   function[i] = func_avg;  i++;
  operators[i] = "runstd";   function[i] = func_std;  i++;

  cdoInitialize(argument);

  operatorInputArg("number of timesteps");
  ndates = atoi(operatorArgv()[0]);

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  nvars    = vlistNvars(vlistID1);
  nrecords = vlistNrecs(vlistID1);

  recVarID   = (int *) malloc(nrecords*sizeof(int));
  recLevelID = (int *) malloc(nrecords*sizeof(int));

  vars1 = (FIELD ***) malloc((ndates+1)*sizeof(FIELD **));
  if ( function[operatorID] == func_std )
    vars2 = (FIELD ***) malloc((ndates+1)*sizeof(FIELD **));

  for ( its = 0; its < ndates; its++ )
    {
      vars1[its] = (FIELD **) malloc(nvars*sizeof(FIELD *));
      if ( function[operatorID] == func_std )
	vars2[its] = (FIELD **) malloc(nvars*sizeof(FIELD *));

      for ( varID = 0; varID < nvars; varID++ )
	{
	  gridID   = vlistInqVarGrid(vlistID1, varID);
	  gridsize = gridInqSize(gridID);
	  nlevel   = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
	  missval  = vlistInqVarMissval(vlistID1, varID);

	  vars1[its][varID] = (FIELD *)  malloc(nlevel*sizeof(FIELD));
	  if ( function[operatorID] == func_std )
	    vars2[its][varID] = (FIELD *)  malloc(nlevel*sizeof(FIELD));

	  for ( levelID = 0; levelID < nlevel; levelID++ )
	    {
	      vars1[its][varID][levelID].grid    = gridID;
	      vars1[its][varID][levelID].nmiss   = 0;
	      vars1[its][varID][levelID].missval = missval;
	      vars1[its][varID][levelID].ptr     = (double *) malloc(gridsize*sizeof(double));
	      if ( function[operatorID] == func_std )
		{
		  vars2[its][varID][levelID].grid    = gridID;
		  vars2[its][varID][levelID].nmiss   = 0;
		  vars2[its][varID][levelID].missval = missval;
		  vars2[its][varID][levelID].ptr     = (double *) malloc(gridsize*sizeof(double));
		}
	    }
	}
    }

  for ( tsID = 0; tsID < ndates; tsID++ )
    {
      nrecs = streamInqTimestep(streamID1, tsID);
      if ( nrecs == 0 )
	cdoAbort("File has less then %d timesteps!", ndates);
	
      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);

	  if ( tsID == 0 )
	    {
	      recVarID[recID]   = varID;
	      recLevelID[recID] = levelID;
	    }
	  
	  streamReadRecord(streamID1, vars1[tsID][varID][levelID].ptr, &nmiss);
	  vars1[tsID][varID][levelID].nmiss = nmiss;

	  if ( function[operatorID] == func_std )
	    {
	      farmoq(&vars2[tsID][varID][levelID], vars1[tsID][varID][levelID]);
	      for ( inp = 0; inp < tsID; inp++ )
		{
		  farsumq(&vars2[inp][varID][levelID], vars1[tsID][varID][levelID]);
		  farsum(&vars1[inp][varID][levelID], vars1[tsID][varID][levelID]);
		}
	    }
	  else
	    {
	      for ( inp = 0; inp < tsID; inp++ )
		{
		  farfun(&vars1[inp][varID][levelID], vars1[tsID][varID][levelID], function[operatorID]);
		}
	    }
	}
    }

  otsID = 0;
  while ( TRUE )
    {
      if      ( function[operatorID] == func_mean || function[operatorID] == func_avg )
	for ( varID = 0; varID < nvars; varID++ )
	  {
	    if ( vlistInqVarTime(vlistID1, varID) == TIME_CONSTANT ) continue;
	    nlevel   = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
	    for ( levelID = 0; levelID < nlevel; levelID++ )
	      farcmul(&vars1[0][varID][levelID], 1.0/ndates);
	  }
      else if ( function[operatorID] == func_std )
	for ( varID = 0; varID < nvars; varID++ )
	  {
	    if ( vlistInqVarTime(vlistID1, varID) == TIME_CONSTANT ) continue;
	    nlevel   = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
	    for ( levelID = 0; levelID < nlevel; levelID++ )
	      farcstd(&vars1[0][varID][levelID], vars2[0][varID][levelID], 1.0/ndates);
	  }

      streamDefTimestep(streamID2, otsID++);

      for ( recID = 0; recID < nrecords; recID++ )
	{
	  varID    = recVarID[recID];
	  levelID  = recLevelID[recID];

	  if ( otsID == 1 || vlistInqVarTime(vlistID1, varID) == TIME_VARIABLE )
	    {
	      streamDefRecord(streamID2, varID, levelID);
	      streamWriteRecord(streamID2, vars1[0][varID][levelID].ptr, vars1[0][varID][levelID].nmiss);
	    }
	}

      vars1[ndates] = vars1[0];
      if ( function[operatorID] == func_std )
	vars2[ndates] = vars2[0];

      for ( inp = 0; inp < ndates; inp++ )
	{
	  vars1[inp] = vars1[inp+1];
	  if ( function[operatorID] == func_std )
	    vars2[inp] = vars2[inp+1];
	}

      nrecs = streamInqTimestep(streamID1, tsID);
      if ( nrecs == 0 ) break;

      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  
	  streamReadRecord(streamID1, vars1[ndates-1][varID][levelID].ptr, &nmiss);
	  vars1[ndates-1][varID][levelID].nmiss = nmiss;

	  if ( function[operatorID] == func_std )
	    {
	      for ( inp = 0; inp < ndates-1; inp++ )
		{
		  farsumq(&vars2[inp][varID][levelID], vars1[ndates-1][varID][levelID]);
		  farsum(&vars1[inp][varID][levelID], vars1[ndates-1][varID][levelID]);
		}
	      farmoq(&vars2[ndates-1][varID][levelID], vars1[ndates-1][varID][levelID]);
	    }
	  else
	    {
	      for ( inp = 0; inp < ndates-1; inp++ )
		{
		  farfun(&vars1[inp][varID][levelID], vars1[ndates-1][varID][levelID], function[operatorID]);
		}
	    }
	}

      tsID++;
    }

  for ( its = 0; its < ndates; its++ )
    {
      for ( varID = 0; varID < nvars; varID++ )
	{
	  nlevel = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
	  for ( levelID = 0; levelID < nlevel; levelID++ )
	    {
	      free(vars1[its][varID][levelID].ptr);
	      if ( function[operatorID] == func_std ) free(vars2[its][varID][levelID].ptr);
	    }

	  free(vars1[its][varID]);
	  if ( function[operatorID] == func_std ) free(vars2[its][varID]);
	}
    }

  free(vars1);
  if ( function[operatorID] == func_std ) free(vars2);

  if ( recVarID   ) free(recVarID);
  if ( recLevelID ) free(recLevelID);

  streamClose(streamID2);
  streamClose(streamID1);

  cdoFinish();

  return (0);
}
