/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Set
@Title     = Set
@Section   = Manipulating the header/field
@Arguments = ifile ofile
@Operators = setpartab setcode setvar

@EndModule


@BeginOperator_setpartab

@Title     = Set parameter table
@Parameter = table

@BeginDesciption
Sets the parameter table for all variables.
@EndDesciption

@BeginParameter
@Item = table
STRING  Parameter table file or name
@EndParameter

@EndOperator


@BeginOperator_setcode

@Title     = Set code
@Parameter = code

@BeginDesciption
Sets the code for all variables to the same given value.
@EndDesciption

@BeginParameter
@Item = code
INTEGER Code number
@EndParameter

@EndOperator


@BeginOperator_setvar

@Title     = Set variable name
@Parameter = name

@BeginDesciption
Sets the name of the first variable.
@EndDesciption

@BeginParameter
@Item = name
STRING  Variable name
@EndParameter

@EndOperator

@EndDoc
*/


void *Set(void *argument)
{
  static char func[] = "Set";
  const char *operators[] = {"setpartab", "setcode", "setvar"};
  enum                     {f_setpartab, f_setcode, f_setvar};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int streamID1, streamID2 = CDI_UNDEFID;
  int nrecs, nvars, newval = -1;
  int tsID1, recID, varID, levelID;
  int vlistID1, vlistID2;
  int taxisID1;
  int nmiss;
  int gridsize;
  int tableID = -1;
  char *newname = NULL, *partab = NULL;
  double *array = NULL;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  if ( operatorID == f_setcode )
    {
      operatorInputArg("code");
      newval = atoi(operatorArgv()[0]);
    }
  else if ( operatorID == f_setvar )
    {
      operatorInputArg("variable name");
      newname = operatorArgv()[0];
    }
  else if ( operatorID == f_setpartab )
    {
      operatorInputArg("parameter table");
      partab = operatorArgv()[0];
      tableID = defineTable(partab);
    }

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  taxisID1 = vlistInqTaxis(vlistID1);

  vlistID2 = vlistDuplicate(vlistID1);

  if ( operatorID == f_setcode )
    {
      nvars = vlistNvars(vlistID2);
      for ( varID = 0; varID < nvars; varID++ )
	vlistDefVarCode(vlistID2, varID, newval);
    }
  else if ( operatorID == f_setvar )
    {
      vlistDefVarName(vlistID2, 0, newname);
    }
  else if ( operatorID == f_setpartab )
    {
      nvars = vlistNvars(vlistID2);
      for ( varID = 0; varID < nvars; varID++ )
	vlistDefVarTable(vlistID2, varID, tableID);
    }

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  gridsize = vlistGridsizeMax(vlistID2);
  array = (double *) malloc(gridsize*sizeof(double));

  tsID1 = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID1)) )
    {
      streamDefTimestep(streamID2, tsID1);
	       
      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamDefRecord(streamID2,  varID,  levelID);
	  
	  streamReadRecord(streamID1, array, &nmiss);
	  streamWriteRecord(streamID2, array, nmiss);
	}
      tsID1++;
    }

  streamClose(streamID1);
  streamClose(streamID2);

  if ( array ) free(array);

  cdoFinish();

  return (0);
}
