/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Setgrid
@Title     = Set grid type
@Section   = Manipulating the header/field
@Arguments = ifile ofile
@Operators = setgrid setgridtype

@EndModule


@BeginOperator_setgrid

@Title     = Set grid
@Parameter = grid

@BeginDesciption
Sets the grid information for all variables.
@EndDesciption

@BeginParameter
@Item = grid
STRING  Grid description file or name of the target grid
@EndParameter

@EndOperator


@BeginOperator_setgridtype

@Title     = Set grid type
@Parameter = gridtype

@BeginDesciption
Sets the grid type for all grids to a user given value.
@EndDesciption

@BeginParameter
@Item = gridtype
STRING  New grid type (curvilinear or cell)
@EndParameter

@EndOperator

@EndDoc
*/

void *Setgrid(void *argument)
{
  static char func[] = "Setgrid";
  const char *operators[] = {"setgrid", "setgridtype"};
  enum                     {f_setgrid, f_setgridtype};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int streamID1, streamID2 = CDI_UNDEFID;
  int nrecs;
  int tsID, recID, varID, levelID;
  int vlistID1, vlistID2;
  int taxisID1, taxisID2;
  int vdate, vtime;
  int gridID1, gridID2 = -1;
  int ngrids, index;
  int gridsize, gridtype = -1;
  int nmiss;
  int found;
  char *gridname = NULL;
  double *array = NULL;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, operators);

  switch (operatorID)
    {
    case f_setgrid:
      operatorInputArg("grid description file or name");
      gridID2 = cdoDefineGrid(operatorArgv()[0]);
      break;
    case f_setgridtype:
      operatorInputArg("grid type");
      gridname = operatorArgv()[0];

      if      ( strcmp(gridname, "curvilinear") == 0 ) gridtype = GRID_CURVILINEAR;
      else if ( strcmp(gridname, "cell") == 0 )        gridtype = GRID_CELL;
      else if ( strcmp(gridname, "lonlat") == 0 )      gridtype = GRID_LONLAT;
      else if ( strcmp(gridname, "gaussian") == 0 )    gridtype = GRID_GAUSSIAN;
      else cdoAbort("%s grid unsupported!", gridname);
      break;
    }

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);

  taxisID1 = vlistInqTaxis(vlistID1);
  taxisID2 = taxisDuplicate(taxisID1);

  vlistDefTaxis(vlistID2, taxisID2);

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  switch (operatorID)
    {
    case f_setgrid:
      found = 0;
      ngrids = vlistNgrids(vlistID1);
      for ( index = 0; index < ngrids; index++ )
	{
	  gridID1 = vlistGrid(vlistID1, index);

	  if ( gridInqSize(gridID1) == gridInqSize(gridID2) )
	    {
	      vlistChangeGridIndex(vlistID2, index, gridID2);
	      found++;
	    }
	}
      if ( ! found ) cdoWarning("No grid with %d points found!", gridInqSize(gridID2));
      break;
    case f_setgridtype:
      ngrids = vlistNgrids(vlistID1);
      for ( index = 0; index < ngrids; index++ )
	{
	  gridID1 = vlistGrid(vlistID1, index);

	  if      ( gridtype == GRID_CURVILINEAR ) gridID2 = gridToCurvilinear(gridID1);
	  else if ( gridtype == GRID_CELL )        gridID2 = gridToCell(gridID1);
	  else cdoAbort("%s grid unsupported!", gridname);

	  vlistChangeGridIndex(vlistID2, index, gridID2);
	}
      break;
    }

  streamDefVlist(streamID2, vlistID2);

  gridsize = vlistGridsizeMax(vlistID1);
  array = (double *) malloc(gridsize*sizeof(double));

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      vdate = taxisInqVdate(taxisID1);
      vtime = taxisInqVtime(taxisID1);

      taxisDefVdate(taxisID2, vdate);
      taxisDefVtime(taxisID2, vtime);

      streamDefTimestep(streamID2, tsID);
	       
      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamDefRecord(streamID2,  varID,  levelID);
	  
	  streamReadRecord(streamID1, array, &nmiss);
	  streamWriteRecord(streamID2, array, nmiss);
	}
      tsID++;
    }

  streamClose(streamID1);
  streamClose(streamID2);

  if ( array ) free(array);

  cdoFinish();

  return (0);
}
