/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Setmiss
@Title     = Set missing data
@Section   = Missing values
@Arguments = ifile ofile
@Operators = setmissval setctomiss setmisstoc setrtomiss

@EndModule


@BeginOperator_setmissval

@Title     = Set a new missing value
@Parameter = miss

@BeginDesciption
@IfMan
         / miss   if i(t,x) EQ miss
o(t,x) = 
         \ i(t,x) if i(t,x) NE miss
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
 \mbox{miss} & \mbox{if} \;\; i(t,x) = \mbox{\sl miss}      \\
 i(t,x)      & \mbox{if} \;\; i(t,x) \neq \mbox{\sl miss}   \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = miss
FLOAT  New missing value
@EndParameter

@EndOperator


@BeginOperator_setctomiss

@Title     = Set constant to missing value
@Parameter = c

@BeginDesciption
@IfMan
         / miss   if i(t,x) EQ c
o(t,x) = 
         \ i(t,x) if i(t,x) NE c
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
 \mbox{miss} & \mbox{if} \;\; i(t,x) = \mbox{\sl c}      \\
 i(t,x)      & \mbox{if} \;\; i(t,x) \neq \mbox{\sl c}   \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = c
FLOAT  Constant
@EndParameter

@EndOperator


@BeginOperator_setmisstoc

@Title     = Set missing value to constant
@Parameter = c

@BeginDesciption
@IfMan
         / c      if i(t,x) EQ miss
o(t,x) = 
         \ i(t,x) if i(t,x) NE miss
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
 \mbox{\sl c} & \mbox{if} \;\; i(t,x) = \mbox{miss}     \\
 i(t,x)       & \mbox{if} \;\; i(t,x) \neq \mbox{miss}  \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter
@Item = c
FLOAT  Constant
@EndParameter

@EndOperator


@BeginOperator_setrtomiss

@Title     = Set range to missing value
@Parameter = rmin rmax

@BeginDesciption
@IfMan
         / miss   if i(t,x) GE rmin AND i(t,x) LE rmax
o(t,x) = 
         \ i(t,x) if i(t,x) LT rmin AND i(t,x) GT rmax
@EndifMan
@IfDoc
@BeginMath
o(t,x) = \left\{
\begin{array}{cll}
 \mbox{miss}   & \mbox{if} \;\; i(t,x) \geq \mbox{\sl rmin} \wedge i(t,x) \leq \mbox{\sl rmax}  \\
 i(t,x) & \mbox{if} \;\; i(t,x) < \mbox{\sl rmin} \vee i(t,x) > \mbox{\sl rmax}  \\
\end{array}   \right.
@EndMath
@EndifDoc
@EndDesciption

@BeginParameter rmax
@Item = rmin
FLOAT  Lower bound
@Item = rmax
FLOAT  Upper bound
@EndParameter

@EndOperator

@EndDoc
*/


void *Setmiss(void *argument)
{
  static char func[] = "Setmiss";
  const char *operators[] = {"setmissval", "setctomiss", "setmisstoc", "setrtomiss"};
  enum                     {f_setmissval, f_setctomiss, f_setmisstoc, f_setrtomiss};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int streamID1, streamID2;
  int gridsize;
  int nrecs, recID;
  int nvars;
  int tsID;
  int varID, levelID;
  int vlistID1, vlistID2;
  int nmiss;
  int i;
  double missval, missval2 = 0;
  double rconst = 0, rmin = 0, rmax = 0;
  double *array;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, operators);

  if ( operatorID == f_setmissval )
    {
      operatorInputArg("missing value");
      if ( operatorArgc() < 1 ) cdoAbort("No enough arguments!");
      missval2 = atof(operatorArgv()[0]);
    }
  else if ( operatorID == f_setctomiss || operatorID == f_setmisstoc )
    {
      operatorInputArg("constant");
      if ( operatorArgc() < 1 ) cdoAbort("No enough arguments!");
      rconst = atof(operatorArgv()[0]);
    }
  else
    {
      operatorInputArg("range (min, max)");
      if ( operatorArgc() < 2 ) cdoAbort("No enough arguments!");
      rmin = atof(operatorArgv()[0]);
      rmax = atof(operatorArgv()[1]);
    }

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);

  if ( operatorID == f_setmissval )
    {
      nvars = vlistNvars(vlistID2);
      for ( varID = 0; varID < nvars; varID++ )
	vlistDefVarMissval(vlistID2, varID, missval2);
    }

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  gridsize = vlistGridsizeMax(vlistID1);

  array = (double *) malloc(gridsize*sizeof(double));

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      streamDefTimestep(streamID2, tsID);

      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamReadRecord(streamID1, array, &nmiss);

	  gridsize = gridInqSize(vlistInqVarGrid(vlistID1, varID));
	  missval = vlistInqVarMissval(vlistID1, varID);

	  if ( operatorID == f_setmissval )
	    {
	      nmiss = 0;
	      for ( i = 0; i < gridsize; i++ )
		if ( array[i] == missval  || array[i] == (float)missval ||
		     array[i] == missval2 || array[i] == (float)missval2 )
		  {
		    array[i] = missval2;
		    nmiss++;
		  }
	    }
	  if ( operatorID == f_setctomiss )
	    {
	      for ( i = 0; i < gridsize; i++ )
		if ( array[i] == rconst || array[i] == (float)rconst )
		  {
		    array[i] = missval;
		    nmiss++;
		  }
	    }
	  else if ( operatorID == f_setmisstoc )
	    {
	      nmiss = 0;
	      for ( i = 0; i < gridsize; i++ )
		if ( array[i] == missval || array[i] == (float)missval )
		  {
		    array[i] = rconst;
		  }
	    }
	  else /* setrtomiss */
	    {
	      for ( i = 0; i < gridsize; i++ )
		if ( array[i] >= rmin && array[i] <= rmax )
		  {
		    array[i] = missval;
		    nmiss++;
		  }
	    }

	  streamDefRecord(streamID2, varID, levelID);
	  streamWriteRecord(streamID2, array, nmiss);
	}
      tsID++;
    }

  streamClose(streamID2);
  streamClose(streamID1);

  if ( array ) free(array);

  cdoFinish();

  return (0);
}
