/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>
#include <ctype.h>  /* isdigit */

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Settime
@Title     = Set time
@Section   = Manipulating the header/field
@Arguments = ifile ofile
@Operators = setdate settime setday setmon setyear settunits settaxis setreftime shifttime

@EndModule


@BeginOperator_setdate

@Title     = Set date
@Parameter = date

@BeginDesciption
Sets the date in every timestep to the same given value.
@EndDesciption

@EndOperator


@BeginOperator_settime

@Title     = Set time
@Parameter = time

@BeginDesciption
Sets the time in every timestep to the same given value.
@EndDesciption

@EndOperator


@BeginOperator_setday

@Title     = Set day
@Parameter = day

@BeginDesciption
Sets the day in every timestep to the same given value.
@EndDesciption

@EndOperator


@BeginOperator_setmon

@Title     = Set month
@Parameter = month

@BeginDesciption
Sets the month in every timestep to the same given value.
@EndDesciption

@EndOperator


@BeginOperator_setyear

@Title     = Set year
@Parameter = year

@BeginDesciption
Sets the year in every timestep to the same given value.
@EndDesciption

@EndOperator


@BeginOperator_settunits

@Title     = Set time units
@Parameter = units

@BeginDesciption
Sets the time units.
@EndDesciption

@BeginParameter units
@Item = units
STRING  Base units of the time axis (minute, hour, day, month, year).
@EndParameter

@EndOperator


@BeginOperator_settaxis

@Title     = Set time axis
@Parameter = date time [inc]

@BeginDesciption
Sets the time axis.
@EndDesciption

@BeginParameter date
@Item = date
INTEGER  Start date (e.g. 19780130)
@Item = time
INTEGER  Start time (e.g. 1800)
@Item = inc
STRING   Optional increment (e.g. 12hour) [default: 0hour]
@EndParameter

@EndOperator


@BeginOperator_setreftime

@Title     = Set reference time
@Parameter = date time

@BeginDesciption
Sets the reference time of an relative time axis.
@EndDesciption

@BeginParameter date
@Item = date
INTEGER  Reference date (e.g. 19780130)
@Item = time
INTEGER  Reference time (e.g. 1800)
@EndParameter

@EndOperator


@BeginOperator_shifttime

@Title     = Shift time steps
@Parameter = sval

@BeginDesciption
Shifts all time steps by the parameter sval.
@EndDesciption

@BeginParameter
@Item = sval
STRING   Shift value (e.g. -3hour)
@EndParameter

@EndOperator

@EndDoc
*/

void *Settime(void *argument)
{
  static char func[] = "Settime";
  const char *operators[] = {"setyear", "setmon", "setday", "setdate", "settime", "settunits", "settaxis", "setreftime", "shifttime"};
  enum                     {f_setyear, f_setmon, f_setday, f_setdate, f_settime, f_settunits, f_settaxis, f_setreftime, f_shifttime};
  const char *ostring[]   = {   "year",    "month",  "day",    "date",   "time", "time units", "date,time<,increment>", "date,time", "shift value"};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int streamID1, streamID2 = CDI_UNDEFID;
  int year, mon, day;
  int nrecs, newval = 0;
  int tsID1, recID, varID, levelID;
  int vlistID1, vlistID2;
  int vdate, vtime;
  int sdate = 0, stime = 0;
  int taxisID1, taxisID2 = CDI_UNDEFID;
  int nmiss;
  int gridsize;
  int tunit = TUNIT_DAY;
  int ijulinc = 0, incperiod = 0, incunit = 0;
  double julval = 0;
  double *array = NULL;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, operators);

  operatorInputArg(ostring[operatorID]);

  if ( operatorID == f_settaxis || operatorID == f_setreftime )
    {
      if ( operatorArgc() < 2 ) cdoAbort("Not enough arguments!");
      sdate = atoi(operatorArgv()[0]);
      stime = atoi(operatorArgv()[1]);
      if ( operatorArgc() == 3 )
	{
	  size_t len;
	  char *unit = operatorArgv()[2];
	  incperiod = atoi(unit);
	  while ( isdigit((int) *unit) ) unit++;
	  len = strlen(unit);      
	  if      ( strncmp(unit, "minute", len) == 0 ) { incunit =    60; tunit = TUNIT_MINUTE;}
	  else if ( strncmp(unit, "hour", len)   == 0 ) { incunit =  3600; tunit = TUNIT_HOUR;}
	  else if ( strncmp(unit, "day", len)    == 0 ) { incunit = 86400; tunit = TUNIT_DAY;}
	  else if ( strncmp(unit, "month", len)  == 0 ) { incunit =     1; tunit = TUNIT_MONTH;}
	  else if ( strncmp(unit, "year", len)   == 0 ) { incunit =    12; tunit = TUNIT_YEAR;}
	  else    cdoWarning("time unit >%s< unsupported", unit);
	}
      /* increment in seconds */
      ijulinc = incperiod * incunit;
    }
  else if ( operatorID == f_shifttime )
    {
	  size_t len;
	  char *unit = operatorArgv()[0];
	  incperiod = atoi(unit);
	  if ( unit[0] == '-' ) unit++;
	  while ( isdigit((int) *unit) ) unit++;
	  len = strlen(unit);      
	  if      ( strncmp(unit, "minute", len) == 0 ) { incunit =    60; tunit = TUNIT_MINUTE;}
	  else if ( strncmp(unit, "hour", len)   == 0 ) { incunit =  3600; tunit = TUNIT_HOUR;}
	  else if ( strncmp(unit, "day", len)    == 0 ) { incunit = 86400; tunit = TUNIT_DAY;}
	  else if ( strncmp(unit, "month", len)  == 0 ) { incunit =     1; tunit = TUNIT_MONTH;}
	  else if ( strncmp(unit, "year", len)   == 0 ) { incunit =    12; tunit = TUNIT_YEAR;}
	  else    cdoWarning("time unit >%s< unsupported", unit);

      /* increment in seconds */
      ijulinc = incperiod * incunit;
    }
  else if ( operatorID == f_settunits )
    {
      size_t len;
      char *unit = operatorArgv()[0];
      len = strlen(unit);      
      if      ( strncmp(unit, "minute", len) == 0 ) { tunit = TUNIT_MINUTE;}
      else if ( strncmp(unit, "hour", len)   == 0 ) { tunit = TUNIT_HOUR;}
      else if ( strncmp(unit, "day", len)    == 0 ) { tunit = TUNIT_DAY;}
      else if ( strncmp(unit, "month", len)  == 0 ) { tunit = TUNIT_MONTH;}
      else if ( strncmp(unit, "year", len)   == 0 ) { tunit = TUNIT_YEAR;}
    }
  else
    {
      newval = atoi(operatorArgv()[0]);
    }

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);

  taxisID1 = vlistInqTaxis(vlistID1);

  if ( operatorID == f_setreftime )
    {
      taxisID2 = taxisNew(TAXIS_RELATIVE);
      if ( operatorArgc() != 3 ) tunit = taxisInqTunit(taxisID1);
      taxisDefTunit(taxisID2, tunit);
    }
  else
    taxisID2 = taxisDuplicate(taxisID1);

  if ( operatorID == f_settaxis )
    {
      taxisDefTunit(taxisID2, tunit);
      taxisDefRdate(taxisID2, sdate);
      taxisDefRtime(taxisID2, stime);
      julval = enjulian(sdate, stime);
    }
  else if ( operatorID == f_settunits )
    {
      taxisDefTunit(taxisID2, tunit);
    }
  else if ( operatorID == f_setreftime )
    {
      taxisDefRdate(taxisID2, sdate);
      taxisDefRtime(taxisID2, stime);
    }

  vlistDefTaxis(vlistID2, taxisID2);

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  gridsize = vlistGridsizeMax(vlistID2);
  array = (double *) malloc(gridsize*sizeof(double));

  tsID1 = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID1)) )
    {
      if ( operatorID == f_settaxis )
	{
	  if ( incunit == 1 || incunit == 12 )
	    {
	      vtime = stime;
	      if ( tsID1 == 0 )
		{
		  vdate = sdate;
		}
	      else
		{
		  year = vdate/10000;
		  mon  = (vdate - year*10000) / 100;
		  day  =  vdate - year*10000 - mon*100;
	      
		  mon += ijulinc;

		  while ( mon > 12 )
		    {
		      mon -= 12;
		      year++;
		    }

		  while ( mon < 1 )
		    {
		      mon += 12;
		      year--;
		    }

		  vdate = year*10000 + mon*100 + day;
		}
	    }
	  else
	    {
	      dejulian(julval, &vdate, &vtime);
	      julval += ijulinc;
	    }
	}
      else if ( operatorID == f_shifttime )
	{
	  vdate = taxisInqVdate(taxisID1);
	  vtime = taxisInqVtime(taxisID1);

	  if ( incunit == 1 || incunit == 12 )
	    {
	      year = vdate/10000;
	      mon  = (vdate - year*10000) / 100;
	      day  =  vdate - year*10000 - mon*100;
	      
	      mon += ijulinc;

	      while ( mon > 12 )
		{
		  mon -= 12;
		  year++;
		}

	      while ( mon < 1 )
		{
		  mon += 12;
		  year--;
		}

	      vdate = year*10000 + mon*100 + day;
	    }
	  else
	    {
	      julval = enjulian(vdate, vtime);
	      julval += ijulinc;
	      dejulian(julval, &vdate, &vtime);
	    }
	}
      else if ( operatorID == f_setreftime )
	{
	  vdate = taxisInqVdate(taxisID1);
	  vtime = taxisInqVtime(taxisID1);
	}
      else
	{
	  vdate = taxisInqVdate(taxisID1);
	  vtime = taxisInqVtime(taxisID1);

	  year = vdate/10000;
	  mon  = (vdate - year*10000) / 100;
	  day  =  vdate - year*10000 - mon*100;

	  if ( operatorID == f_setyear ) year = newval;
	  if ( operatorID == f_setmon  ) mon  = newval;
	  if ( operatorID == f_setday  ) day  = newval;
      
	  vdate = year*10000 + mon*100 + day;

	  if ( operatorID == f_setdate  ) vdate = newval;
	  if ( operatorID == f_settime  ) vtime = newval;
	}

      taxisDefVdate(taxisID2, vdate);
      taxisDefVtime(taxisID2, vtime);

      streamDefTimestep(streamID2, tsID1);
	       
      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamDefRecord(streamID2,  varID,  levelID);
	  
	  streamReadRecord(streamID1, array, &nmiss);
	  streamWriteRecord(streamID2, array, nmiss);
	}
      tsID1++;
    }

  streamClose(streamID1);
  streamClose(streamID2);

  if ( array ) free(array);

  cdoFinish();

  return (0);
}
