/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <stdio.h>
#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Showinfo
@Title     = 
@Section   = Information
@Arguments = ifile
@Operators = showyear showmon showdate showtime showcode showvar showlevel

@EndModule


@BeginOperator_showyear

@Title     = Show years

@BeginDesciption
Prints all different years.
@EndDesciption

@EndOperator


@BeginOperator_showmon

@Title     = Show months

@BeginDesciption
Prints all different months.
@EndDesciption

@EndOperator


@BeginOperator_showdate

@Title     = Show dates

@BeginDesciption
Prints all different dates.
@EndDesciption

@EndOperator


@BeginOperator_showtime

@Title     = Show timesteps

@BeginDesciption
Prints all timesteps.
@EndDesciption

@EndOperator


@BeginOperator_showcode

@Title     = Show codes

@BeginDesciption
Prints the code of all different variables.
@EndDesciption

@EndOperator


@BeginOperator_showvar

@Title     = Show variable names

@BeginDesciption
Print all different variable names.
@EndDesciption

@EndOperator


@BeginOperator_showlevel

@Title     = Show levels

@BeginDesciption
Prints all levels for each variable.
@EndDesciption

@EndOperator

@EndDoc
*/

void *Showinfo(void *argument)
{
  const char *operators[] = {"showyear", "showmon", "showdate", "showtime", "showcode", "showvar", "showlevel"};
  enum                     {f_showyear, f_showmon, f_showdate, f_showtime, f_showcode, f_showvar, f_showlevel};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int varID, zaxisID;
  int vdate, vtime;
  int nrecs, nvars, ntsteps, nout;
  int nlevs, levelID;
  int tsID, ndate, date0 = 0;
  int mon0 = 0, mon, nmon, year0 = 0, year, nyear;
  int taxisID;
  int streamID;
  int vlistID;
  char varname[128];

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, operators);

  streamID = streamOpenRead(cdoStreamName(0));
  if ( streamID < 0 ) cdiError(streamID, "Open failed on %s", cdoStreamName(0));

  vlistID = streamInqVlist(streamID);

  nvars   = vlistNvars(vlistID);
  taxisID = vlistInqTaxis(vlistID);
  ntsteps = vlistNtsteps(vlistID);

  switch (operatorID)
    {
    case f_showyear:
      nyear = 0;
      tsID = 0;
      while ( (nrecs = streamInqTimestep(streamID, tsID)) )
	{
	  vdate = taxisInqVdate(taxisID);

	  year = vdate/10000;
	 
	  if ( tsID == 0 || year0 != year )
	    {
	      if ( nyear == 10 ) { nyear = 0; fprintf(stdout, "\n"); }
	      year0 = year;
	      fprintf(stdout, " %4d", year0);
	      nyear++;
	    }

	  tsID++;
	}
      fprintf(stdout, "\n");
      break;
    case f_showmon:
      nmon = 0;
      tsID = 0;
      while ( (nrecs = streamInqTimestep(streamID, tsID)) )
	{
	  vdate = taxisInqVdate(taxisID);

	  year = vdate/10000;
	  mon  = (vdate - year*10000) / 100;
	 
	  if ( tsID == 0 || mon0 != mon )
	    {
	      if ( nmon == 12 ) { nmon = 0; fprintf(stdout, "\n"); }
	      mon0 = mon;
	      fprintf(stdout, " %2d", mon0);
	      nmon++;
	    }

	  tsID++;
	}
      fprintf(stdout, "\n");
      break;
    case f_showdate:
      ndate = 0;
      tsID  = 0;
      while ( (nrecs = streamInqTimestep(streamID, tsID)) )
	{
	  vdate = taxisInqVdate(taxisID);
	 
	  if ( tsID == 0 || date0 != vdate )
	    {
	      if ( ndate == 10 ) { ndate = 0; fprintf(stdout, "\n"); }
	      date0 = vdate;
	      fprintf(stdout, " %d", date0);
	      ndate++;
	    }

	  tsID++;
	}
      fprintf(stdout, "\n");
      break;
    case f_showtime:
      nout = 0;
      tsID = 0;
      while ( (nrecs = streamInqTimestep(streamID, tsID)) )
	{
	  if ( nout == 4 ) { nout = 0; fprintf(stdout, "\n"); }
	  vdate = taxisInqVdate(taxisID);
	  vtime = taxisInqVtime(taxisID);
	  fprintf(stdout, " %d-%d", vdate, vtime);
	  tsID++;
	  nout++;
	}
      fprintf(stdout, "\n");
      break;
    case f_showcode:
      nout = 0;
      for ( varID = 0; varID < nvars; varID++ )
	{
	  if ( nout == 20 ) { nout = 0; fprintf(stdout, "\n"); }
	  fprintf(stdout, " %d", vlistInqVarCode(vlistID, varID));
	  nout++;
	}
      fprintf(stdout, "\n");
      break;
    case f_showvar:
      for ( varID = 0; varID < nvars; varID++ )
	{
	  vlistInqVarName(vlistID, varID, varname);
	  fprintf(stdout, " %s", varname);
	  fprintf(stdout, "\n");
	}
      break;
    case f_showlevel:
      for ( varID = 0; varID < nvars; varID++ )
	{
	  zaxisID = vlistInqVarZaxis(vlistID, varID);
	  nlevs = zaxisInqSize(zaxisID);
	  for ( levelID = 0; levelID < nlevs; levelID++ )
	    fprintf(stdout, " %.9g", zaxisInqLevel(zaxisID, levelID));
	  fprintf(stdout, "\n");
	}
      break;
    default:
      cdoAbort("operator not implemented!");
    }

  streamClose(streamID);

  cdoFinish();

  return (0);
}
