/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <stdio.h>
#include <string.h>
#include "pipe.h"
#include "pstream_int.h"
#include "cdi.h"
#include "cdo.h"
#include "error.h"
#include "dmemory.h"

#if  defined  (HAVE_PTHREAD_H)

static int PipeDebug = 0;

void pipeInit(PIPE *pipe)
{
  static char func[] = "pipeInit";
  pipe->EOP  = 0;

  pipe->recIDr  = -1;
  pipe->recIDw  = -1;
  pipe->tsIDr   = -1;
  pipe->tsIDw   = -1;

  pipe->nmiss   = 0;
  pipe->data    = NULL;
  pipe->hasdata = 0;
  pipe->usedata = TRUE;
  pipe->pstreamIDin = -1;

  pipe->mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(pipe->mutex, NULL);

  pipe->tsDef = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));
  pthread_cond_init(pipe->tsDef, NULL);
  pipe->tsInq = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));
  pthread_cond_init(pipe->tsInq, NULL);

  pipe->recDef = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));
  pthread_cond_init(pipe->recDef, NULL);
  pipe->recInq = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));
  pthread_cond_init(pipe->recInq, NULL);
  
  pipe->vlistDef = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));
  pthread_cond_init(pipe->vlistDef, NULL);
  pipe->isclosed = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));
  pthread_cond_init(pipe->isclosed, NULL);

  pipe->writeCond = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));
  pthread_cond_init(pipe->writeCond, NULL);

  pipe->readCond = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));
  pthread_cond_init(pipe->readCond, NULL);
}

void pipeDefVlist(int pstreamID, int vlistID)
{
  static char func[] = "pipeDefVlist";
  char *pname = pstreamTable[pstreamID].name;
  PIPE *pipe;

  if ( PipeDebug )
    Message(func, "%s pstreamID %d", pname, pstreamID);

  pipe = pstreamTable[pstreamID].pipe;

  pthread_mutex_lock(pipe->mutex);
  pstreamTable[pstreamID].vlistID = vlistID;
  pthread_mutex_unlock(pipe->mutex);  

  pthread_cond_signal(pipe->vlistDef);
}

int pipeInqVlist(int pstreamID)
{
  static char func[] = "pipeInqVlist";
  char *pname = pstreamTable[pstreamID].name;
  PIPE *pipe;
  int vlistID;

  if ( PipeDebug )
    Message(func, "%s pstreamID %d", pname, pstreamID);

  pipe = pstreamTable[pstreamID].pipe;

  pthread_mutex_lock(pipe->mutex);
  while ( pstreamTable[pstreamID].vlistID == -1 )
    {
      if ( PipeDebug ) Message(func, "%s wait of vlist", pname);
      pthread_cond_wait(pipe->vlistDef, pipe->mutex);
    }
  vlistID = pstreamTable[pstreamID].vlistID;
  pthread_mutex_unlock(pipe->mutex);

  return (vlistID);
}

int pipeInqTimestep(int pstreamID, int tsID)
{
  static char func[] = "pipeInqTimestep";
  char *pname = pstreamTable[pstreamID].name;
  PIPE *pipe;
  int nrecs;

  if ( PipeDebug )
    Message(func, "%s pstreamID %d", pname, pstreamID);

  pipe = pstreamTable[pstreamID].pipe;

  pthread_mutex_lock(pipe->mutex);
  pipe->usedata = FALSE;
  pipe->recIDr = -1;
  if ( tsID != pipe->tsIDr+1 )
    {
      if ( ! (tsID == pipe->tsIDr && pipe->tsIDr == pipe->tsIDw && pipe->recIDr == -1) ) 
	Error(func, "%s unexpected tsID %d %d %d", pname, tsID, pipe->tsIDr+1, pipe->tsIDw);
    }
  
  pipe->tsIDr = tsID;
  while ( pipe->tsIDw != tsID )
    {
      if ( pipe->EOP == TRUE )
	{
	  if ( PipeDebug ) Message(func, "%s EOP", pname);
	  break;
	}
      if ( pipe->hasdata )
	{
	  if ( PipeDebug ) Message(func, "%s has data", pname);
	  pipe->hasdata = 0;
	  pipe->data = NULL;
	  pthread_cond_signal(pipe->readCond);
	}
      else
	if ( PipeDebug ) Message(func, "%s has no data", pname);

      pthread_cond_signal(pipe->recInq); /* o.k. ??? */

      if ( PipeDebug ) Message(func, "%s wait of tsDef", pname);
      pthread_cond_wait(pipe->tsDef, pipe->mutex);
    }

  if ( pipe->EOP == TRUE )
    nrecs = 0;
  else
    nrecs = pipe->nrecs;

  pthread_mutex_unlock(pipe->mutex);

  pthread_cond_signal(pipe->tsInq);

  return (nrecs);
}

void pipeDefTimestep(int pstreamID, int tsID)
{
  static char func[] = "pipeDefTimestep";
  char *pname = pstreamTable[pstreamID].name;
  int nrecs;
  PIPE *pipe;

  if ( PipeDebug )
    Message(func, "%s pstreamID %d", pname, pstreamID);

  pipe = pstreamTable[pstreamID].pipe;

  pthread_mutex_lock(pipe->mutex);
  pipe->recIDw = -1;
  pipe->tsIDw++;
  if ( tsID != pipe->tsIDw )
    Error(func, "unexpected tsID %d(%d) for %s", tsID, pipe->tsIDw, pname);

  if ( tsID == 0 )
    nrecs = vlistNrecs(pstreamTable[pstreamID].vlistID);
  else
    {
      int vlistID, varID;
      vlistID = pstreamTable[pstreamID].vlistID;
      nrecs = 0;
      for ( varID = 0; varID < vlistNvars(vlistID); varID++ )
	if ( vlistInqVarTime(vlistID, varID) == TIME_VARIABLE )
	  nrecs += zaxisInqSize(vlistInqVarZaxis(vlistID, varID));
    }

  pipe->nrecs = nrecs;
  pthread_mutex_unlock(pipe->mutex);
  pthread_cond_signal(pipe->tsDef);
  /*
sleep(1);
*/
  pthread_mutex_lock(pipe->mutex);
  while ( pipe->tsIDr < tsID )
    {
      if ( pipe->EOP == TRUE )
	{
	  if ( PipeDebug ) Message(func, "EOP");
	  break;
	}
      if ( PipeDebug ) Message(func, "%s wait of tsInq", pname);
      pthread_cond_wait(pipe->tsInq, pipe->mutex);
    }
  pthread_mutex_unlock(pipe->mutex);
}

int pipeInqRecord(int pstreamID, int *varID, int *levelID)
{
  static char func[] = "pipeInqRecord";
  char *pname = pstreamTable[pstreamID].name;
  PIPE *pipe;
  int condSignal = FALSE;

  if ( PipeDebug )
    Message(func, "%s pstreamID %d", pname, pstreamID);
 
  pipe = pstreamTable[pstreamID].pipe;

  pthread_mutex_lock(pipe->mutex);

  if ( PipeDebug )
    Message(func, "%s has no data %d %d", pname, pipe->recIDr, pipe->recIDw);

  if ( pipe->hasdata || pipe->usedata )
    {
      pipe->hasdata = 0;
      pipe->data = NULL;
      pipe->usedata = FALSE;
      condSignal = TRUE;
    }
 	  
  pthread_mutex_unlock(pipe->mutex);
  if ( condSignal ) pthread_cond_signal(pipe->readCond);

  pthread_mutex_lock(pipe->mutex);

  pipe->usedata = TRUE;

  pipe->recIDr++;
  
  if ( PipeDebug )
    Message(func, "%s recID %d %d", pname, pipe->recIDr, pipe->recIDw);  

  while ( pipe->recIDw != pipe->recIDr )
    {
      if ( pipe->EOP == TRUE )
	{
	  if ( PipeDebug ) Message(func, "EOP");
	  break;
	}
      if ( PipeDebug ) Message(func, "%s wait of recDef", pname);
      pthread_cond_wait(pipe->recDef, pipe->mutex);
    }

  if ( pipe->EOP == TRUE )
    {
      *varID   = -1;
      *levelID = -1;
    }
  else
    {
      *varID   = pipe->varID;
      *levelID = pipe->levelID;
    }

  pthread_mutex_unlock(pipe->mutex);

  pthread_cond_signal(pipe->recInq);

  return (0);
}

void pipeDefRecord(int pstreamID, int varID, int levelID)
{
  static char func[] = "pipeDefRecord";
  char *pname = pstreamTable[pstreamID].name;
  PIPE *pipe;
  int condSignal = FALSE;

  if ( PipeDebug )
    Message(func, "%s pstreamID %d", pname, pstreamID);

  pipe = pstreamTable[pstreamID].pipe;

  pthread_mutex_lock(pipe->mutex);

  if ( PipeDebug )
    Message(func, "%s has data %d %d", pname, pipe->recIDr, pipe->recIDw);
	  
  if ( pipe->hasdata )
    {
      pipe->hasdata = 0;
      pipe->data = NULL;
      condSignal = TRUE;
    }

  pthread_mutex_unlock(pipe->mutex);
  if ( condSignal ) pthread_cond_signal(pipe->readCond);

  pthread_mutex_lock(pipe->mutex);
  pipe->usedata = TRUE;
  pipe->recIDw++;

  if ( PipeDebug )
    Message(func, "%s recID %d %d", pname, pipe->recIDr, pipe->recIDw);  

  pipe->varID = varID;
  pipe->levelID = levelID;
  pthread_mutex_unlock(pipe->mutex);
  pthread_cond_signal(pipe->recDef);

  pthread_mutex_lock(pipe->mutex);
  while ( pipe->recIDr < pipe->recIDw )
    {
      if ( pipe->tsIDw != pipe->tsIDr ) break;
      if ( PipeDebug ) Message(func, "%s wait of recInq %d", pname, pipe->recIDr);
      pthread_cond_wait(pipe->recInq, pipe->mutex);
    }
  pthread_mutex_unlock(pipe->mutex);
}

void pipeCopyRecord(int pstreamIDout, int pstreamIDin)
{
  static char func[] = "pipeCopyRecord";
  char *ipname = pstreamTable[pstreamIDin].name;
  char *opname = pstreamTable[pstreamIDout].name;
  PIPE *pipe;

  if ( PipeDebug )
    Message(func, "%s pstreamIDin %d", ipname, pstreamIDin);
  if ( PipeDebug )
    Message(func, "%s pstreamIDout %d", opname, pstreamIDout);

  pipe = pstreamTable[pstreamIDout].pipe;

  pthread_mutex_lock(pipe->mutex);
  pipe->hasdata = 2; /* pipe */
  pipe->pstreamIDin = pstreamIDin;
  pthread_mutex_unlock(pipe->mutex);

  pthread_cond_signal(pipe->writeCond);

  pthread_mutex_lock(pipe->mutex);
  while ( pipe->hasdata )
    {
      if ( pipe->usedata == FALSE ) break;

      if ( pipe->recIDw != pipe->recIDr ) break;

      if ( pipe->EOP == TRUE )
	{
	  if ( PipeDebug ) Message(func, "EOP");
	  break;
	}
      if ( PipeDebug ) Message(func, "%s wait of read record", opname);
      pthread_cond_wait(pipe->readCond, pipe->mutex);
    }
  pthread_mutex_unlock(pipe->mutex);
}

void pipeReadRecord(int pstreamID, double *data, int *nmiss)
{
  static char func[] = "pipeReadRecord";
  char *pname = pstreamTable[pstreamID].name;
  PIPE *pipe;

  *nmiss = 0;
  if ( PipeDebug )
    Message(func, "%s pstreamID %d", pname, pstreamID);

  pipe = pstreamTable[pstreamID].pipe;

  pthread_mutex_lock(pipe->mutex);
  while ( pipe->hasdata == 0 )
    {
      if ( PipeDebug ) Message(func, "%s wait of writeCond", pname);
      pthread_cond_wait(pipe->writeCond, pipe->mutex);
    }

  if ( pipe->hasdata == 2 )
    {
      int pstreamIDin;

      pstreamIDin = pipe->pstreamIDin;
      /*
      if ( pstreamTable[pstreamIDin].ispipe )
	{
	  pstreamIDin = pstreamTable[pstreamIDin].pipe->pstreamIDin;
	  if ( pstreamTable[pstreamIDin].ispipe )
	    Error(func, "istream is pipe");
	}
      */
      pstreamID   = pstreamIDin;
      while ( pstreamTable[pstreamIDin].ispipe )
	{
	  if ( PipeDebug ) fprintf(stderr, "%s: istream %d is pipe\n", func, pstreamIDin);
	  pstreamID   = pstreamIDin;
	  pstreamIDin = pstreamTable[pstreamIDin].pipe->pstreamIDin;
	  if ( pstreamIDin == -1 ) break;
	}

      if ( pstreamIDin == -1 )
	{
	  if ( PipeDebug ) fprintf(stderr, "pstreamID = %d\n", pstreamID);
	  if ( pstreamTable[pstreamID].pipe->hasdata == 1 )
	    {
	      int vlistID, gridsize;

	      if ( ! pstreamTable[pstreamID].pipe->data )
		Error(func, "No data pointer for %s", pname);

	      vlistID = pstreamTable[pstreamID].vlistID;
	      gridsize = gridInqSize(vlistInqVarGrid(vlistID, pstreamTable[pstreamID].pipe->varID));
	      memcpy(data, pstreamTable[pstreamID].pipe->data, gridsize*sizeof(double));
	      *nmiss = pstreamTable[pstreamID].pipe->nmiss;
	    }
	  else
	    Error(func, "Internal problem! istream undefined");
	}
      else
	{
	  if ( PipeDebug ) fprintf(stderr, "%s: istream %d is file\n", func, pstreamIDin);
	  streamReadRecord(pstreamTable[pstreamIDin].fileID, data, nmiss);
	}
    }
  else if ( pipe->hasdata == 1 )
    {
      int vlistID, gridsize;

      if ( ! pipe->data )
	Error(func, "No data pointer for %s", pname);

      vlistID = pstreamTable[pstreamID].vlistID;
      gridsize = gridInqSize(vlistInqVarGrid(vlistID, pipe->varID));
      memcpy(data, pipe->data, gridsize*sizeof(double));
      *nmiss = pipe->nmiss;
    }
  else
    {
      Error(func, "data type %d not implemented", pipe->hasdata);
    }

  if ( PipeDebug ) Message(func, "%s read record %d", pname, pipe->recIDr);
 

  pipe->hasdata = 0;
  pipe->data = NULL;
  pthread_mutex_unlock(pipe->mutex);

  pthread_cond_signal(pipe->readCond);
}

void pipeWriteRecord(int pstreamID, double *data, int nmiss)
{
  static char func[] = "pipeWriteRecord";
  char *pname = pstreamTable[pstreamID].name;
  PIPE *pipe;

  if ( PipeDebug )
    Message(func, "%s pstreamID %d", pname, pstreamID);

  pipe = pstreamTable[pstreamID].pipe;
  /*
  if ( ! pipe->usedata ) return;
  */
  pthread_mutex_lock(pipe->mutex);
  pipe->hasdata = 1; /* data pointer */
  pipe->data    = data;
  pipe->nmiss   = nmiss;
  pthread_mutex_unlock(pipe->mutex);
  pthread_cond_signal(pipe->writeCond);

  if ( PipeDebug ) Message(func, "%s write record %d", pname, pipe->recIDw);

  pthread_mutex_lock(pipe->mutex);
  while ( pipe->hasdata )
    {
      if ( pipe->usedata == FALSE ) break;
      /*     
      printf("ts ids %d %d\n", pipe->tsIDw, pipe->tsIDr);
      printf("rec ids %d %d\n", pipe->recIDw, pipe->recIDr);
      */
      if ( pipe->recIDw != pipe->recIDr ) break;

      if ( pipe->EOP == TRUE )
	{
	  if ( PipeDebug ) Message(func, "EOP");
	  break;
	}
      if ( PipeDebug ) Message(func, "%s wait of readCond", pname);
      pthread_cond_wait(pipe->readCond, pipe->mutex);
    }
  pthread_mutex_unlock(pipe->mutex);
}

void pipeDebug(int debug)
{
  PipeDebug = debug;
}

#endif
