/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "error.h"

#define UNDEFID -1

typedef struct {
  double *vals;
  int     type;
  size_t  size;
  char    name[128];
  char    units[128];
}
ZAXIS;

void zaxisInit(ZAXIS *zaxis)
{
  zaxis->vals        = NULL;
  zaxis->type        = UNDEFID;
  zaxis->size        = 0;
  zaxis->name[0]     = 0;
  zaxis->units[0]    = 0;
}

static int getoptname(char *optname, const char *optstring, int nopt)
{
  int i, nerr = 0;
  size_t namelen;
  const char *pname;
  const char *pend;

  pname = optstring;
  pend  = optstring;

  for ( i = 0; i < nopt; i++ )
    {
      pend = strchr(pname, ',');
      if ( pend == NULL )
	break;
      else
	pname = pend + 1;
    }

  if ( pend )
    {
      pend = strchr(pname, ',');
      if ( pend == NULL )
	namelen = strlen(pname);
      else
	namelen = pend - pname;

      strncpy(optname, pname, namelen);
      optname[namelen] = '\0';
    }
  else
    nerr = 1;

  return (nerr);
}

int zaxisDefine(ZAXIS zaxis)
{
  static char func[] = "zaxisDefine";
  int zaxisID = UNDEFID;

  if ( zaxis.type == -1 ) Error(func, "zaxistype undefined!");

  if ( zaxis.size == 0 ) Error(func, "zaxis size undefined");

  zaxisID = zaxisNew(zaxis.type, zaxis.size);

  if ( zaxis.vals )     zaxisDefLevels(zaxisID, zaxis.vals);
  if ( zaxis.name[0] )  zaxisDefName(zaxisID, zaxis.name);
  if ( zaxis.units[0] ) zaxisDefUnits(zaxisID, zaxis.units);

  return (zaxisID);
}

static char *skipSeparator(char *pline)
{
  while ( isspace((int) *pline) ) pline++;
  if ( *pline == '=' || *pline == ':' ) pline++;
  while ( isspace((int) *pline) ) pline++;

  return (pline);
}

int zaxisFromFile(FILE *gfp)
{
  static char func[] = "zaxisFromFile";
  char line[1024], *pline;
  int zaxisID;
  ZAXIS zaxis;

  zaxisInit(&zaxis);

  while ( readline(gfp, line, 1024) )
    {
      if ( line[0] == '#' ) continue;
      if ( line[0] == '\0' ) continue;
      pline = line;
      while ( isspace((int) *pline) ) pline++;
      if ( pline[0] == '\0' ) continue;
      if ( strncmp(pline, "type", 4) == 0 )
	{
	  pline = skipSeparator(pline + 4);
	  if ( strncmp(pline, "pressure", 6) == 0 )
	    zaxis.type = ZAXIS_PRESSURE;
	  else if ( strncmp(pline, "hybrid", 6)  == 0 )
	    zaxis.type = ZAXIS_HYBRID;
	  else if ( strncmp(pline, "height", 6)  == 0 )
	    zaxis.type = ZAXIS_HEIGHT;
	  else if ( strncmp(pline, "depth below sea", 15)  == 0 )
	    zaxis.type = ZAXIS_DEPTH_BELOW_SEA;
	  else if ( strncmp(pline, "depth below land", 16)  == 0 )
	    zaxis.type = ZAXIS_DEPTH_BELOW_LAND;
	  else if ( strncmp(pline, "isentropic", 10)  == 0 )
	    zaxis.type = ZAXIS_ISENTROPIC;
	  else
	    Warning(func, "Invalid zaxisname : %s", pline);
	}
      else if ( strncmp(pline, "nlev", 4)  == 0 )
	{
	  zaxis.size = atol(skipSeparator(pline + 4));
	}
      else if ( strncmp(pline, "name", 4)  == 0 )
	{
	  strcpy(zaxis.name, skipSeparator(pline + 4));
	}
      else if ( strncmp(pline, "units", 5)  == 0 )
	{
	  strcpy(zaxis.units, skipSeparator(pline + 5));
	}
      else if ( strncmp(pline, "levels", 6)  == 0 )
	{
	  int i;
	  float flev;

	  pline = skipSeparator(pline + 6);
	  
	  zaxis.vals = (double *) malloc(zaxis.size*sizeof(double));
	  for ( i = 0; i < (int) zaxis.size; i++ )
	    {
	      pline = skipSeparator(pline);
	      if ( strlen(pline) == 0 )
		{
		  if ( ! readline(gfp, line, 1024) )
		    {
		      Warning(func, "Incomplete commmand: >levels<");
		      break;
		    }
		  pline = line;
		  pline = skipSeparator(pline);
		}
	      sscanf(pline, "%g", &flev);
	      zaxis.vals[i] = flev;
	      while ( isalnum((int) *pline) ||
		      isdigit((int) *pline) ||
		      ispunct((int) *pline) ) pline++;
	    }
	}
      else
	Warning(func, "Invalid zaxis command : >%s<", pline);
    }

  zaxisID = zaxisDefine(zaxis);

  return (zaxisID);
}

void defineZaxis(const char *zaxisarg)
{
  static char func[] = "defineZaxis";
  char zaxisfile[1024];
  FILE *zfp;
  int nfile = 0;

  while ( getoptname(zaxisfile, zaxisarg, nfile++) == 0 )
    {      
      zfp = fopen(zaxisfile, "r");
      if ( zfp == 0 )
	{
	  SysError(func, zaxisfile);
	}
      else
	{
	  (void) zaxisFromFile(zfp);
	  fclose(zfp);
	}
    }
}
